#!/usr/bin/env python
# -*- coding: utf-8 -*-

# tools.stats.py
"""
Take numerical data from main app and convert to a format suitable for statistical analysis
in Regression and Control Chart tabs

"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import numpy as np
from scipy import stats as scipy_stats
from sklearn import linear_model
from sklearn.metrics import mean_squared_error, r2_score
from regressors import stats as regressors_stats
from dvha.db import sql_columns


class StatsData:
    def __init__(self, dvhs, table_data, group=1):
        """
        Class used to to collect data for Regression and Control Chart
        This process is different than for Time Series since regressions require all variables to be the same length
        :param dvhs: data from DVH query
        :type dvhs: DVH
        :param table_data: table data other than from DVHs
        :type table_data: dict
        """
        self.dvhs = dvhs
        self.table_data = table_data
        self.group = group

        self.column_info = sql_columns.numerical
        self.correlation_variables = list(self.column_info)
        self.correlation_variables.sort()

        self.__map_data()

    def __map_data(self):
        self.data = {}
        stat_types = ['min', 'mean', 'median', 'max']
        for var in self.correlation_variables:
            if var in self.column_info.keys():
                var_name = self.column_info[var]['var_name']
                table = self.column_info[var]['table']

                if table == 'DVHs':
                    self.data[var] = {'units': self.column_info[var]['units'],
                                      'values': getattr(self.dvhs, var_name)}

                # single value variables
                elif table == 'Plans':
                    src = self.table_data[table]
                    self.data[var] = {'units': self.column_info[var]['units'],
                                      'values': [getattr(src, var_name)[self.get_plan_index(uid)] for uid in self.uids]}

                # multi value variables
                elif table == 'Beams':
                    src = self.table_data[table]

                    if str_starts_with_any_in_list(var, ['Beam Complexity', 'Beam Area', 'Control Point MU',
                                                         'Beam Perimeter', 'Beam Energy']):
                        # stats of these four variable types have min, mean, median, and max types in DB
                        # The following will take min, mean, median, or max of all values for a UID based on var type
                        # Example, if var_name == Beam Complexity (Max), the following will return the Max of these
                        temp = []
                        for uid in self.uids:
                            indices = self.get_beam_indices(uid)
                            beam_data = getattr(self.table_data['Beams'], var_name)
                            values = [beam_data[i] for i in indices if beam_data[i] != 'None']
                            for stat in stat_types:
                                if stat in var.lower():
                                    if values:
                                        temp.append(getattr(np, stat)(values))
                                    else:
                                        temp.append(None)
                        self.data[var] = {'units': self.column_info[var]['units'],
                                          'values': temp}
                    else:
                        temp = {s: [] for s in stat_types}
                        for uid in self.uids:
                            for stat in stat_types:
                                values = self.get_src_values(src, var_name, uid)
                                values = [v for v in values if v != 'None']
                                if values:
                                    temp[stat].append(getattr(np, stat)(values))
                                else:
                                    temp[stat].append(None)

                        for stat in stat_types:
                            corr_key = "%s (%s)" % (var, stat.capitalize())
                            self.data[corr_key] = {'units': self.column_info[var]['units'],
                                                   'values': temp[stat]}
        self.validate_data()

    def validate_data(self):
        """
        Remove any variables that are constant to avoid crash on regression
        """
        bad_vars = []
        for var_name, var_obj in self.data.items():
            if 'Date' in var_name:
                if var_name != 'Simulation Date':
                    bad_vars.append(var_name)
            else:
                values = [float(val) for val in var_obj['values'] if val != 'None' and val is not None]
                if not any(np.diff(values).tolist()):
                    bad_vars.append(var_name)

        for var in bad_vars:
            self.data.pop(var)

    def update_endpoints_and_radbio(self):
        """
        Update endpoint and radbio data in self.data. This function is needed since all of these values are calculated
        after a query and user may change these values.
        """
        if self.dvhs:
            if self.dvhs.endpoints['defs']:
                for var in self.dvhs.endpoints['defs']['label']:
                    if var not in self.variables:
                        self.data[var] = {'units': '',
                                          'values': self.dvhs.endpoints['data'][var]}

                for var in self.variables:
                    if var[0:2] in {'D_', 'V_'}:
                        if var not in self.dvhs.endpoints['defs']['label']:
                            self.data.pop(var)

            if self.dvhs.eud:
                self.data['EUD'] = {'units': 'Gy',
                                    'values': self.dvhs.eud}
            if self.dvhs.ntcp_or_tcp:
                self.data['NTCP or TCP'] = {'units': '',
                                            'values': self.dvhs.ntcp_or_tcp}
            self.validate_data()

    @staticmethod
    def get_src_values(src, var_name, uid):
        uid_indices = [i for i, x in enumerate(src.study_instance_uid) if x == uid]
        return [getattr(src, var_name)[i] for i in uid_indices]

    def get_plan_index(self, uid):
        return self.table_data['Plans'].study_instance_uid.index(uid)

    def get_beam_indices(self, uid):
        return [i for i, x in enumerate(self.table_data['Beams'].study_instance_uid) if x == uid]

    def get_bokeh_data(self, x, y):
        """
        Get data in a format compatible with bokeh's ColumnDataSource.data
        :param x: x-variable name
        :type x: str
        :param y: y-variable name
        :type y: str
        :return: x and y data
        :rtype: dict
        """
        if x in list(self.data) and y in list(self.data):
            # TODO: potential data length issue with study_instance_uid
            # Received the following error with 0.6.9, can't reproduce
            # BokehUserWarning: ColumnDataSource's columns must be of the same length. Current lengths:
            # ('date', 69), ('mrn', 69), ('uid', 70), ('x', 69), ('y', 69)
            # Appears to point to this function's return?
            return {'uid': self.uids,
                    'mrn': self.mrns,
                    'date': self.sim_study_dates,
                    'x': self.data[x]['values'],
                    'y': self.data[y]['values']}
        return {key: [] for key in ['uid', 'mrn', 'date', 'x', 'y']}

    @property
    def uids(self):
        return self.dvhs.study_instance_uid

    @property
    def mrns(self):
        return self.dvhs.mrn

    @property
    def sim_study_dates(self):
        return self.data['Simulation Date']['values']

    @property
    def variables(self):
        return [var for var in list(self.data) if var != 'Simulation Date']

    @property
    def trending_variables(self):
        return list(self.data)

    def get_axis_title(self, variable):
        if self.data[variable]['units']:
            return "%s (%s)" % (variable, self.data[variable]['units'])
        return variable

    def get_X_and_y(self, y_variable, x_variables, include_patient_info=False):
        """
        Collect data for input into multi-variable regression
        :param y_variable: dependent variable
        :type y_variable: str
        :param x_variables: independent variables
        :type x_variables: list
        :param include_patient_info: If True, return mrn, uid, dates with X and y
        :type include_patient_info: bool
        :return: X, y or X, y, mrn, uid, dates
        """
        data, mrn, uid, dates = [], [], [], []
        y_var_data = []
        for i, value in enumerate(self.data[y_variable]['values']):
            y_var_data.append([value, np.nan][value == 'None'])
            mrn.append(self.mrns[i])
            uid.append(self.uids[i])
            dates.append(self.sim_study_dates[i])

        data.append(y_var_data)
        for var in x_variables:
            x_var_data = []
            for value in self.data[var]['values']:
                x_var_data.append([value, np.nan][value == 'None'])
            data.append(x_var_data)

        data = np.array(data)
        bad_indices = get_index_of_nan(data)

        for bad_index in bad_indices[::-1]:
            data = np.delete(data, bad_index, 1)
            mrn.pop(bad_index)
            uid.pop(bad_index)
            dates.pop(bad_index)

        X = np.transpose(data[1:])
        y = data[0]

        if not include_patient_info:
            return X, y
        return X, y, mrn, uid, dates

    def add_variable(self, variable, values, units=''):
        if variable not in list(self.data):
            self.data[variable] = {'units': units, 'values': values}
        if variable not in self.correlation_variables:
            self.correlation_variables.append(variable)
            self.correlation_variables.sort()

    def del_variable(self, variable):
        if variable in list(self.data):
            self.data.pop(variable)
        if variable in self.correlation_variables:
            index = self.correlation_variables.index(variable)
            self.correlation_variables.pop(index)

    def set_variable_data(self, variable, data, units=None):
        self.data[variable]['values'] = data
        if units is not None:
            self.data[variable]['units'] = units

    def set_variable_units(self, variable, units):
        self.data[variable]['units'] = units

    def get_corr_matrix_data(self, options, included_vars=None, extra_vars=None):
        if included_vars is None:
            included_vars = list(self.data)
        if extra_vars is not None:
            included_vars = included_vars + extra_vars
        else:
            extra_vars = []

        categories = [c for c in list(self.data) if 'date' not in c.lower() and c in included_vars]
        categories.extend(extra_vars)
        categories = list(set(categories))
        categories.sort()
        var_count = len(categories)
        categories_for_label = [category.replace("Control Point", "CP") for category in categories]
        categories_for_label = [category.replace("control point", "CP") for category in categories_for_label]
        categories_for_label = [category.replace("Distance", "Dist") for category in categories_for_label]

        for i, category in enumerate(categories_for_label):
            if category.startswith('DVH'):
                categories_for_label[i] = category.split("DVH Endpoint: ")[1]

        x_factors = categories_for_label
        y_factors = categories_for_label[::-1]

        s_keys = ['x', 'y', 'x_name', 'y_name', 'color', 'alpha', 'r', 'p', 'size',
                  'x_normality', 'y_normality', 'group']
        source_data = {'corr': {sk: [] for sk in s_keys},
                       'line': {'x': [0.5, var_count - 0.5], 'y': [var_count - 0.5, 0.5]}}

        min_size, max_size = 3, 20
        for x in range(var_count):
            for y in range(var_count):
                if x > y and self.group == 1 or x < y and self.group == 2:
                    if categories[x] not in extra_vars and categories[y] not in extra_vars:

                        bad_indices = [i for i, v in enumerate(self.data[categories[x]]['values']) if type(v) is str]
                        bad_indices.extend([i for i, v in enumerate(self.data[categories[y]]['values']) if type(v) is str])
                        bad_indices = list(set(bad_indices))

                        x_data = [v for i, v in enumerate(self.data[categories[x]]['values']) if i not in bad_indices]
                        y_data = [v for i, v in enumerate(self.data[categories[y]]['values']) if i not in bad_indices]

                        if x_data and len(x_data) == len(y_data):
                            r, p_value = scipy_stats.pearsonr(x_data, y_data)
                        else:
                            r, p_value = 0, 0
                        if np.isnan(r):
                            r = 0

                        sign = ['neg', 'pos'][r >= 0]
                        color = getattr(options, 'CORRELATION_%s_COLOR_%s' % (sign.upper(), self.group))
                        source_data['corr']['color'].append(color)
                        source_data['corr']['r'].append(r)
                        source_data['corr']['p'].append(p_value)
                        source_data['corr']['alpha'].append(abs(r))
                        source_data['corr']['size'].append(((max_size - min_size) * abs(r)) + min_size)
                        source_data['corr']['x'].append(x + 0.5)  # 0.5 offset due to bokeh 0.12.9 bug
                        source_data['corr']['y'].append(var_count - y - 0.5)  # 0.5 offset due to bokeh 0.12.9 bug
                        source_data['corr']['x_name'].append(categories_for_label[x])
                        source_data['corr']['y_name'].append(categories_for_label[y])
                        source_data['corr']['group'].append(self.group)

                        try:
                            x_norm, x_p = scipy_stats.normaltest(x_data)
                        except ValueError:
                            x_p = 'N/A'
                        try:
                            y_norm, y_p = scipy_stats.normaltest(y_data)
                        except ValueError:
                            y_p = 'N/A'

                        source_data['corr']['x_normality'].append(x_p)
                        source_data['corr']['y_normality'].append(y_p)

        return {'source_data': source_data, 'x_factors': x_factors, 'y_factors': y_factors}


def get_index_of_nan(numpy_array):
    bad_indices = []
    nan_data = np.isnan(numpy_array)
    for var_data in nan_data:
        indices = np.where(var_data)[0].tolist()
        if indices:
            bad_indices.extend(indices)
    bad_indices = list(set(bad_indices))
    bad_indices.sort()
    return bad_indices


def str_starts_with_any_in_list(string_a, string_list):
    """
    Check if string_a starts with any string the provided list of strings
    """
    for string_b in string_list:
        if string_a.startswith(string_b):
            return True
    return False


def get_p_values(X, y, predictions, params):
    """
    Get p-values using sklearn
    based on https://stackoverflow.com/questions/27928275/find-p-value-significance-in-scikit-learn-linearregression
    :param X: independent data
    :type X: np.array
    :param y: dependent data
    :type y: np.array
    :param predictions: output from linear_model.LinearRegression.predict
    :param params: np.array([y_incercept, slope])
    :return: p-values
    :rtype: list
    """

    newX = np.append(np.ones((len(X), 1)), X, axis=1)
    MSE = (sum((y - predictions) ** 2)) / (len(newX) - len(newX[0]))

    var_b = MSE * (np.linalg.inv(np.dot(newX.T, newX)).diagonal())
    sd_b = np.sqrt(var_b)
    ts_b = params / sd_b

    return [2 * (1 - scipy_stats.t.cdf(np.abs(i), (len(newX) - 1))) for i in ts_b], sd_b, ts_b


class MultiVariableRegression:
    """
    Perform a multi-variable regression using sklearn
    """
    def __init__(self, X, y, saved_reg=None):
        """
        :param X: independent data
        :type X: np.array
        :param y: dependent data
        :type y: list
        """

        if saved_reg is None:
            self.reg = linear_model.LinearRegression()
            self.ols = self.reg.fit(X, y)
        else:
            self.reg = saved_reg.reg
            self.ols = saved_reg.ols

        self.y_intercept = self.reg.intercept_
        self.slope = self.reg.coef_
        params = np.append(self.y_intercept, self.slope)
        self.predictions = self.reg.predict(X)

        self.r_sq = r2_score(y,  self.predictions)
        self.mse = mean_squared_error(y,  self.predictions)

        self.p_values, self.sd_b, self.ts_b = get_p_values(X, y,  self.predictions, params)

        self.residuals = np.subtract(y, self.predictions)

        self.norm_prob_plot = scipy_stats.probplot(self.residuals, dist='norm', fit=False, plot=None, rvalue=False)

        reg_prob = linear_model.LinearRegression()
        reg_prob.fit([[val] for val in self.norm_prob_plot[0]], self.norm_prob_plot[1])

        self.y_intercept_prob = reg_prob.intercept_
        self.slope_prob = reg_prob.coef_
        self.x_trend_prob = [min(self.norm_prob_plot[0]), max(self.norm_prob_plot[0])]
        self.y_trend_prob = np.add(np.multiply(self.x_trend_prob, self.slope_prob), self.y_intercept_prob)

        self.f_stat = regressors_stats.f_stat(self.ols, X, y)
        self.df_error = len(X[:, 0]) - len(X[0, :]) - 1
        self.df_model = len(X[0, :])

        self.f_p_value = scipy_stats.f.cdf(self.f_stat, self.df_model, self.df_error)


def get_control_limits(y, std_devs=3):
    """
    Calculate control limits for Control Chart
    :param y: data
    :type y: list
    :param std_devs: values greater than std_devs away are out-of-control
    :type std_devs: int or float
    :return: center line, upper control limit, and lower control limit
    """
    y = np.array(y)

    center_line = np.mean(y)
    avg_moving_range = np.mean(np.absolute(np.diff(y)))

    scalar_d = 1.128  # since moving range is calculated based on 2 consecutive points

    ucl = center_line + std_devs * avg_moving_range / scalar_d
    lcl = center_line - std_devs * avg_moving_range / scalar_d

    return center_line, ucl, lcl


def sync_variables_in_stats_data_objects(stats_data_1, stats_data_2):
    """
    Ensure both stats_data objects have the same variables
    :type stats_data_1: StatsData
    :type stats_data_2: StatsData
    """

    stats_data = {1: stats_data_1, 2: stats_data_2}
    variables = {grp: list(sd.data) for grp, sd in stats_data.items()}

    for grp, sd in stats_data.items():
        for var in variables[grp]:
            if var not in variables[3-grp]:
                values = ['None'] * len(stats_data[3-grp].mrns)
                units = stats_data[grp].data[var]['units']
                stats_data[3-grp].add_variable(var, values, units)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# db.sql_connector.py
"""
Tools used to communicate with the SQL database
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import psycopg2
import sqlite3
from datetime import datetime
from dateutil.parser import parse as date_parser
from os.path import dirname, join, isfile
from dvha.options import Options
from dvha.paths import CREATE_PGSQL_TABLES, CREATE_SQLITE_TABLES, DATA_DIR
from dvha.tools.errors import SQLError


class DVH_SQL:
    """
    This class is used to communicate to the SQL database to limit the need for syntax in other files
    """
    def __init__(self, *config, db_type='pgsql'):
        """
        :param config: optional SQL login credentials, stored values used if nothing provided
        :param db_type: either 'pgsql' or 'sqlite'
        :type db_type: str
        """

        stored_options = Options()

        if config:
            self.db_type = db_type
            config = config[0]
        else:
            # Read SQL configuration file
            self.db_type = stored_options.DB_TYPE
            config = stored_options.SQL_LAST_CNX[self.db_type]

        if self.db_type == 'sqlite':
            db_file_path = config['host']
            if not dirname(db_file_path):  # file_path has not directory, assume it lives in DATA_DIR
                db_file_path = join(DATA_DIR, db_file_path)
            self.db_name = None
            self.cnx = sqlite3.connect(db_file_path)
        else:
            self.db_name = config['dbname']
            self.cnx = psycopg2.connect(**config)

        self.cursor = self.cnx.cursor()
        self.tables = ['DVHs', 'Plans', 'Rxs', 'Beams', 'DICOM_Files']

    def __enter__(self):
        return self

    def __exit__(self, ctx_type, ctx_value, ctx_traceback):
        self.close()

    def close(self):
        """
        Close the SQL DB connection
        """
        self.cnx.close()

    def execute_file(self, sql_file_name):
        """
        Executes lines within provided text file to SQL
        :param sql_file_name: absolute file path of a text file containing SQL commands
        :type sql_file_name: str
        """

        for line in open(sql_file_name):
            if not line.startswith('--'):  # ignore commented lines
                self.cursor.execute(line)
        self.cnx.commit()

    def execute_str(self, command_str):
        """
        Execute and commit a string in proper SQL syntax, can handle multiple lines split by \n
        :param command_str: command or commands to be executed and committed
        :type command_str: str
        """
        for line in command_str.split('\n'):
            if line:
                self.cursor.execute(line)
        self.cnx.commit()

    def check_table_exists(self, table_name):
        """
        :param table_name: the SQL table to check
        :return: existence of specified table
        :rtype: bool
        """

        self.cursor.execute("""
            SELECT COUNT(*)
            FROM information_schema.tables
            WHERE table_name = '{0}'
            """.format(table_name.replace('\'', '\'\'')))
        return self.cursor.fetchone()[0] == 1

    def query(self, table_name, return_col_str, *condition_str, **kwargs):
        """
        A generalized query function for DVHA
        :param table_name: 'DVHs', 'Plans', 'Rxs', 'Beams', or 'DICOM_Files'
        :type table_name: str
        :param return_col_str: a csv of SQL columns to be returned
        :type return_col_str: str
        :param condition_str: a condition in SQL syntax
        :type condition_str: str
        :param kwargs: optional parameters order, order_by, and bokeh_cds
        :return: results of the query

        kwargs:
            order: specify order direction (ASC or DESC)
            order_by: the column order is applied to
            bokeh_cds: structure data into a format readily accepted by bokeh's ColumnDataSource.data
        """
        order, order_by = None, None
        if kwargs:
            if 'order' in kwargs:
                order = kwargs['order']
            if 'order_by' in kwargs:
                order_by = kwargs['order_by']
                if not order:
                    order = 'ASC'

        query = "Select %s from %s;" % (return_col_str, table_name)
        if condition_str and condition_str[0]:
            query = "Select %s from %s where %s;" % (return_col_str, table_name, condition_str[0])
        if order and order_by:
            query = "%s Order By %s %s;" % (query[:-1], order_by, order)

        try:
            self.cursor.execute(query)
            results = self.cursor.fetchall()
        except Exception as e:
            raise SQLError(str(e), query)

        if 'bokeh_cds' in kwargs and kwargs['bokeh_cds']:
            keys = [c.strip() for c in return_col_str.split(',')]
            results = {key: [results[r][i] for r in range(len(results))] for i, key in enumerate(keys)}
            for key in list(results):
                if 'time' in key or 'date' in key:
                    results[key] = [str(value) for value in results[key]]

        return results

    def query_generic(self, query_str):
        """
        A generic query function that executes the provided string
        :param query_str: SQL command
        :type query_str: str
        :return: query results
        """
        self.cursor.execute(query_str)
        return self.cursor.fetchall()

    @property
    def now(self):
        """
        This function is useful for store a timestamp to be used for deleting all data since this return
        For example, a user canceling an import
        :return: The current time as seen by the SQL database
        :rtype: datetime
        """
        return self.query_generic("SELECT %s" % self.sql_cmd_now)[0][0]

    @property
    def sql_cmd_now(self):
        if self.db_type == 'sqlite':
            sql_cmd = "strftime('%Y-%m-%d %H:%M:%f', 'now', 'localtime')"
        else:
            sql_cmd = "NOW()"
        return sql_cmd

    def update(self, table_name, column, value, condition_str):
        """
        Change the data in the database.
        :param table_name: 'DVHs', 'Plans', 'Rxs', 'Beams', or 'DICOM_Files'
        :type table_name: str
        :param column: SQL column to be updated
        :type column: str
        :param value: value to be set
        :type value: str
        :param condition_str: a condition in SQL syntax
        :type condition_str: str
        """

        try:
            float(value)
            value_is_numeric = True
        except ValueError:
            value_is_numeric = False

        if '::date' in str(value):
            amend_type = ['', '::date'][self.db_type == 'pgsql']  # sqlite3 does not support ::date
            value = "'%s'%s" % (value.strip('::date'), amend_type)  # augment value for postgresql date formatting
        elif value_is_numeric:
            value = str(value)
        elif 'null' == str(value.lower()):
            value = "NULL"
        else:
            value = "'%s'" % str(value)  # need quotes to input a string

        update = "Update %s SET %s = %s WHERE %s" % (table_name, column, value, condition_str)

        try:
            self.cursor.execute(update)
            self.cnx.commit()
        except Exception as e:
            raise SQLError(str(e), update)

    def is_study_instance_uid_in_table(self, table_name, study_instance_uid):
        # As of DVH v0.7.5, study_instance_uid may end with _N where N is the nth plan of a file set
        query = "SELECT DISTINCT study_instance_uid FROM %s WHERE study_instance_uid LIKE '%s%%';" % \
                (table_name, study_instance_uid)
        self.cursor.execute(query)
        results = self.cursor.fetchall()
        return bool(results)

    def is_mrn_in_table(self, table_name, mrn):
        return self.is_value_in_table(table_name, mrn, 'mrn')

    def is_value_in_table(self, table_name, value, column):
        query = "Select Distinct %s from %s where %s = '%s';" % (column, table_name, column, value)
        self.cursor.execute(query)
        results = self.cursor.fetchall()
        return bool(results)

    def insert_row(self, table, row):
        """
        Generic function to import data to the database
        :param table: SQL table name
        :type table: str
        :param row: data returned from DICOM_Parser.get_<table>_row()
        """
        columns = list(row)

        values = []
        for column in columns:
            if row[column] is None or row[column][0] is None or row[column][0] == '':
                if column == 'import_time_stamp':
                    values.append(self.sql_cmd_now)
                else:
                    values.append("NULL")
            else:
                value = row[column][0]
                value_type = row[column][1]

                if 'varchar' in value_type:
                    max_length = int(value_type.replace('varchar(', '').replace(')', ''))
                    values.append("'%s'" % truncate_string(value, max_length))

                elif value_type in {'time_stamp', 'date'}:
                    date = date_parser(value)
                    value = date.date()
                    if value_type == 'time_stamp':
                        value = "%s %s" % (value, date.time())
                    if self.db_type == 'pgsql':  # sqlite3 does not support ::date
                        values.append("'%s'::date" % value)
                    else:
                        values.append("'%s'" % value)

                else:
                    values.append("'%s'" % value)

        cmd = "INSERT INTO %s (%s) VALUES (%s);\n" % (table, ','.join(columns), ",".join(values))
        self.execute_str(cmd)

    def insert_data_set(self, data_set):
        """
        Insert an entire data set for a plan
        :param data_set: a dication of data with table names for keys, and a list of row data for values
        :type data_set: dict
        """
        for key in list(data_set):
            for row in data_set[key]:
                self.insert_row(key, row)

    def get_dicom_file_paths(self, mrn=None, uid=None):
        """
        Lookup the dicom file paths of imported data
        :param mrn: MRN
        :type mrn: str
        :param uid: study instance uid
        :type uid: str
        :return: dicom_file_paths as returned from self.query()
        """
        condition = None
        if uid:
            condition = "study_instance_uid = '%s'" % uid
        elif mrn:
            condition = "mrn = '%s'" % mrn

        if condition is not None:
            columns = 'mrn, study_instance_uid, folder_path, plan_file, structure_file, dose_file'
            return self.query('DICOM_Files', columns, condition, bokeh_cds=True)

    def delete_rows(self, condition_str, ignore_tables=None):
        """
        Delete all rows from all tables not in ignore_table for a given condition. Useful when deleting a plan/patient
        :param condition_str: a condition in SQL syntax
        :type condition_str: str
        :param ignore_tables: tables to be excluded from row deletions
        :type ignore_tables: list
        """

        tables = set(self.tables)
        if ignore_tables:
            tables = tables - set(ignore_tables)

        for table in tables:
            self.cursor.execute("DELETE FROM %s WHERE %s;" % (table, condition_str))
            self.cnx.commit()

    def change_mrn(self, old, new):
        """
        Edit all mrns in database
        :param old: current mrn
        :type old: str
        :param new: new mrn
        :type new: str
        """
        condition = "mrn = '%s'" % old
        for table in self.tables:
            self.update(table, 'mrn', new, condition)

    def change_uid(self, old, new):
        """
        Edit study instance uids in database
        :param old: current study instance uid
        :type old: str
        :param new: new study instance uid
        :type new: str
        """
        condition = "study_instance_uid = '%s'" % old
        for table in self.tables:
            self.update(table, 'study_instance_uid', new, condition)

    def delete_dvh(self, roi_name, study_instance_uid):
        """
        Delete a specified DVHs table row
        :param roi_name: the roi name for the row to be deleted
        :type roi_name: str
        :param study_instance_uid: the associated study instance uid
        :type study_instance_uid: str
        """
        self.cursor.execute("DELETE FROM DVHs WHERE roi_name = '%s' and study_instance_uid = '%s';"
                            % (roi_name, study_instance_uid))
        self.cnx.commit()

    def ignore_dvh(self, variation, study_instance_uid, unignore=False):
        """
        Change an uncategorized roi name to ignored so that it won't show up in the list of uncategorized rois, so that
        the user doesn't have to evaluate its need everytime they cleanup the misc rois imported
        :param variation: roi name
        :type variation: str
        :param study_instance_uid: the associated study instance uid
        :type study_instance_uid: str
        :param unignore: if set to True, sets the variation to 'uncategorized'
        :type unignore: bool
        """
        physician_roi = ['ignored', 'uncategorized'][unignore]
        self.update('dvhs', 'physician_roi', physician_roi,
                    "roi_name = '%s' and study_instance_uid = '%s'" % (variation, study_instance_uid))

    def drop_tables(self):
        """Delete all tables in the database if they exist"""
        for table in self.tables:
            self.cursor.execute("DROP TABLE IF EXISTS %s;" % table)
            self.cnx.commit()

    def drop_table(self, table):
        """
        Delete a table in the database if it exists
        :param table: SQL table
        :type table: str
        """
        self.cursor.execute("DROP TABLE IF EXISTS %s;" % table)
        self.cnx.commit()

    def initialize_database(self):
        """Ensure that all of the latest SQL columns exist in the user's database"""
        create_tables_file = [CREATE_PGSQL_TABLES, CREATE_SQLITE_TABLES][self.db_type == 'sqlite']
        self.execute_file(create_tables_file)

    def reinitialize_database(self):
        """Delete all data and create all tables with latest columns"""
        self.drop_tables()
        self.vacuum()
        self.initialize_database()

    def vacuum(self):
        """Call to reclaim space in the database"""
        if self.db_type == 'sqlite':
            self.cnx.isolation_level = None
            self.cnx.execute('VACUUM')
            self.cnx.isolation_level = ''
        else:
            # TODO: PGSQL VACUUM needs testing
            # self.cnx.execute('VACUUM')
            pass

    def does_db_exist(self):
        """
        Check if database exists
        :return: existence of database
        :rtype: bool
        """
        if self.db_name:
            line = "SELECT datname FROM pg_catalog.pg_database WHERE lower(datname) = lower('%s');" % self.db_name
            self.cursor.execute(line)

            return bool(len(self.cursor.fetchone()))
        else:
            return True

    def is_sql_table_empty(self, table):
        """
        Check if specifed SQL table is empty
        :param table: SQL table
        :type table: str
        :return: True if specified table has no data
        :rtype: bool
        """
        line = "SELECT COUNT(*) FROM %s;" % table
        self.cursor.execute(line)
        count = self.cursor.fetchone()[0]
        return not(bool(count))

    def get_unique_values(self, table, column, *condition, **kwargs):
        """
        Uses SELECT DISTINCT to get distinct values in database
        :param table: SQL table
        :type table: str
        :param column: SQL column
        :type column: str
        :param condition: optional condition in SQL syntax
        :type condition: str
        :param kwargs: option to ignore null values in return
        :return: unique values from database, sorted alphabetically
        :rtype: list
        """
        if condition and condition[0]:
            query = "select distinct %s from %s where %s;" % (column, table, str(condition[0]))
        else:
            query = "select distinct %s from %s;" % (column, table)
        self.cursor.execute(query)
        cursor_return = self.cursor.fetchall()
        if 'ignore_null' in kwargs and kwargs['ignore_null']:
            unique_values = [str(uv[0]) for uv in cursor_return if str(uv[0])]
        else:
            unique_values = [str(uv[0]) for uv in cursor_return]

        unique_values.sort()
        return unique_values

    def get_column_names(self, table_name):
        """
        Get all of the column names for a specified table
        :param table_name: SQL table
        :type table_name: str
        :return: column names of specified table, sorted alphabetically
        :rtype: list
        """
        if self.db_type == 'sqlite':
            query = "PRAGMA table_info(%s);" % table_name.lower()
            index = 1
        else:
            query = "select column_name from information_schema.columns where table_name = '%s';" % table_name.lower()
            index = 0
        self.cursor.execute(query)
        cursor_return = self.cursor.fetchall()
        columns = [str(c[index]) for c in cursor_return]
        columns.sort()
        return columns

    def is_sqlite_column_datetime(self, table_name, column):
        if self.db_type == 'sqlite':
            query = "PRAGMA table_info(%s);" % table_name.lower()
            self.cursor.execute(query)
            cursor_return = self.cursor.fetchall()
            columns = {str(c[1]): str(c[2]) for c in cursor_return}
            if column in list(columns):
                column_type = columns[column]
                return 'time' in column_type.lower() or 'date' in column_type.lower()
        return False

    def get_min_value(self, table, column, condition=None):
        """
        Get the minimum value in the database for a given table and column
        :param table: SQL table
        :type table: str
        :param column: SQL column
        :type column: str
        :param condition: optional condition in SQL syntax
        :type condition: str
        :return: single minimum value in database
        """
        return self.get_sql_function_value('MIN', table, column, condition=condition)

    def get_max_value(self, table, column, condition=None):
        """
        Get the maximum value in the database for a given table and column
        :param table: SQL table
        :type table: str
        :param column: SQL column
        :type column: str
        :param condition: optional condition in SQL syntax
        :type condition: str
        :return: single maximum value in database
        """
        return self.get_sql_function_value('MAX', table, column, condition=condition)

    def get_sql_function_value(self, func, table, column, condition=None, first_value_only=True):
        """
        Used by get_min_values and get_max_values
        :param func: SQL compatible function
        :param table: SQL table
        :type table: str
        :param column: SQL column
        :type column: str
        :param condition: optional condition in SQL syntax
        :type condition: str
        :param first_value_only: if true, only return the first value, otherwise all values returned
        :type first_value_only: bool
        :return: value returned by specified function
        """
        if condition:
            query = "SELECT %s(%s) FROM %s WHERE %s;" % (func, column, table, condition)
        else:
            query = "SELECT %s(%s) FROM %s;" % (func, column, table)
        self.cursor.execute(query)
        cursor_return = self.cursor.fetchone()
        if first_value_only:
            return cursor_return[0]
        return cursor_return

    def get_roi_count_from_query(self, uid=None, dvh_condition=None):
        """
        Counts the DVH rows that match the provided conditions
        :param uid: study instance uid
        :type uid: str
        :param dvh_condition: condition in SQL syntax for the DVHs table
        :type dvh_condition: str
        :return: number of DVH rows
        :rtype: int
        """
        if uid:
            condition = "study_instance_uid in ('%s')" % "', '".join(dvh_condition)
            if dvh_condition:
                condition = " and " + condition
        else:
            condition = ''

        if dvh_condition:
            condition = dvh_condition + condition

        return len(self.query('DVHs', 'mrn', condition))

    def is_uid_imported(self, uid):
        """
        Check all tables to see if study instance uid is used
        :param uid: study instance uid
        :type uid: str
        :return: True if study instance uid exists in any table
        :rtype: bool
        """
        for table in self.tables:
            if self.is_study_instance_uid_in_table(table, uid):
                return True
        return False

    def is_mrn_imported(self, mrn):
        """
        Check all tables to see if MRN is used
        :param mrn: MRN
        :type mrn: str
        :return: True if MRN exists in any table
        :rtype: bool
        """
        for table in self.tables:
            if self.is_mrn_in_table(table, mrn):
                return True
        return False

    def is_roi_imported(self, roi_name, study_instance_uid):
        """
        Check if a study is already using a specified roi name
        :param roi_name: roi name to check
        :type roi_name: str
        :param study_instance_uid: restrict search to this study_instance_uid
        :type study_instance_uid: str
        :return: True if the roi name provided has been used in the specified study
        :rtype: bool
        """
        condition = "roi_name = '%s' and study_instance_uid = '%s'" % (roi_name, study_instance_uid)
        roi_names = self.get_unique_values('DVHs', 'roi_name', condition)
        return bool(roi_names)


def truncate_string(input_string, character_limit):
    """
    Used to truncate a string to ensure it may be imported into database
    :param input_string: string to be truncated
    :type input_string: str
    :param character_limit: the maximum number of allowed characters
    :type character_limit: int
    :return: truncated string (removing the trailing characters)
    :rtype: str
    """
    if len(input_string) > character_limit:
        return input_string[0:(character_limit-1)]
    return input_string


def echo_sql_db(config=None, db_type='pgsql'):
    """
    Echo the database using stored or provided credentials
    :param config: database login credentials
    :type config: dict
    :param db_type: either 'pgsql' or 'sqlite'
    :type db_type: str
    :return: True if connection could be established
    :rtype: bool
    """
    try:
        if config:
            if db_type == 'pgsql' and ('dbname' not in list(config) or 'port' not in list(config)):
                return False
            cnx = DVH_SQL(config, db_type=db_type)
        else:
            cnx = DVH_SQL()
        cnx.close()
        return True
    except Exception as e:
        if type(e) not in [psycopg2.OperationalError, sqlite3.OperationalError]:
            print(str(e))
        return False


def initialize_db():
    with DVH_SQL() as cnx:
        cnx.initialize_database()


def is_file_sqlite_db(sqlite_db_file):
    if isfile(sqlite_db_file):
        try:
            cnx = sqlite3.connect(sqlite_db_file)
            cnx.close()
            return True
        except sqlite3.OperationalError:
            pass

    return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# tools.utilties.py
"""
General utilities for DVHA

"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import wx
from datetime import datetime
from dateutil.parser import parse as parse_date
import linecache
import numpy as np
from os import walk, listdir, unlink, mkdir, rmdir, chdir, sep
from os.path import join, isfile, isdir, splitext, basename, dirname, realpath
import pickle
import pydicom
from pydicom.uid import ImplicitVRLittleEndian
import shutil
from subprocess import check_output
import sys
import tracemalloc
from dvha.db.sql_connector import DVH_SQL
from dvha.paths import SQL_CNF_PATH, WIN_APP_ICON, PIP_LIST_PATH, DIRECTORIES


IGNORED_FILES = ['.ds_store']


def is_windows():
    return wx.Platform == '__WXMSW__'


def set_msw_background_color(window_obj, color='lightgrey'):
    if is_windows():
        window_obj.SetBackgroundColour(color)


def is_linux():
    return wx.Platform == '__WXGTK__'


def is_mac():
    return wx.Platform == '__WXMAC__'


def initialize_directories():
    """Based on paths.py, create required directories if they do not exist"""
    for directory in DIRECTORIES.values():
        if not isdir(directory):
            mkdir(directory)
    initialize_protocols()


def initialize_protocols():
    """If no protocols are found in PROTOCOL_DIR, copy defaults"""
    current_protocols = [f for f in listdir(DIRECTORIES['PROTOCOL']) if splitext(f)[1].lower() == '.scp']
    if not current_protocols:
        file_names = [f for f in listdir(DIRECTORIES['PROTOCOL_DEFAULT']) if splitext(f)[1].lower() == '.scp']
        file_paths = [join(DIRECTORIES['PROTOCOL_DEFAULT'], f) for f in file_names]
        move_files_to_new_path(file_paths, DIRECTORIES['PROTOCOL'], copy_files=True)


def write_sql_connection_settings(config):
    """
    Create a file storing the SQL login credentials
    :param config: contains values for 'host', 'dbname', 'port' and optionally 'user' and 'password'
    :type config: dict
    """
    # TODO: Make this more secure

    text = ["%s %s" % (key, value) for key, value in config.items() if value]
    text = '\n'.join(text)

    with open(SQL_CNF_PATH, "w") as text_file:
        text_file.write(text)


def scale_bitmap(bitmap, width, height):
    """
    Used to scale tool bar images for MSW and GTK, MAC automatically scales
    :param bitmap: bitmap to be scaled
    type bitmap: Bitmap
    :param width: width of output bitmap
    :type width: int
    :param height: height of output bitmap
    :type height: int
    :return: scaled bitmap
    :rtype: Bitmap
    """
    image = wx.Bitmap.ConvertToImage(bitmap)
    image = image.Scale(width, height, wx.IMAGE_QUALITY_HIGH)
    return wx.Bitmap(image)


def get_tree_ctrl_image(file_path, file_type=wx.BITMAP_TYPE_PNG, width=16, height=16):
    """
    Create an image top be used in the TreeCtrl from the provided file_path
    :param file_path: absolute file_path of image
    :type file_path: str
    :param file_type: specify the image format (PNG by default)
    :param width: width of output bitmap (16 default)
    :type width: int
    :param height: height of output bitmap (16 default)
    :type height: int
    :return: scaled image for TreeCtrl
    :rtype: Image
    """
    return wx.Image(file_path, file_type).Scale(width, height).ConvertToBitmap()


def get_file_paths(start_path, search_subfolders=False, extension=None):
    """
    Get a list of absolute file paths for a given directory
    :param start_path: initial directory
    :type start_path str
    :param search_subfolders: optionally search all sub folders
    :type search_subfolders: bool
    :param extension: optionally include only files with specified extension
    :type extension: str
    :return: absolute file paths
    :rtype: list
    """
    if isdir(start_path):
        if search_subfolders:
            file_paths = []
            for root, dirs, files in walk(start_path, topdown=False):
                for name in files:
                    if extension is None or splitext(name)[1].lower() == extension.lower():
                        if name.lower() not in IGNORED_FILES:
                            file_paths.append(join(root, name))
            return file_paths

        file_paths = []
        for f in listdir(start_path):
            if isfile(join(start_path, f)):
                if extension is None or splitext(f)[1].lower() == extension.lower():
                    if f.lower() not in IGNORED_FILES:
                        file_paths.append(join(start_path, f))
        return file_paths
    return []


def get_study_instance_uids(**kwargs):
    """
    Get lists of study instance uids in the SQL database that meet provided conditions
    The values return in the 'common' key are used for the DVH class in models.dvh.py
    :param kwargs: keys are SQL table names and the values are conditions in SQL syntax
    :return: study instance uids for each table, uids found in all tables, and a list of unique uids
    :rtype: dict
    """
    with DVH_SQL() as cnx:
        uids = {table: cnx.get_unique_values(table, 'study_instance_uid', condition)
                for table, condition in kwargs.items()}

    complete_list = flatten_list_of_lists(list(uids.values()), remove_duplicates=True)

    uids['common'] = [uid for uid in complete_list if is_uid_in_all_keys(uid, uids)]
    uids['unique'] = complete_list

    return uids


def is_uid_in_all_keys(uid, uids):
    """
    Check if uid is found in each of the uid lists for each SQL table
    :param uid: study instance uid
    :type uid: str
    :param uids: lists of study instance uids organized by SQL table
    :type uids: dict
    :return: True only if uid is found in each of the tables
    :rtype: bool
    """

    table_answer = {}
    # Initialize a False value for each key
    for table in list(uids):
        table_answer[table] = False
    # search for uid in each keyword fof uid_kwlist
    for table, value in uids.items():
        if uid in value:
            table_answer[table] = True

    final_answer = True
    # Product of all answer[key] values (except 'unique')
    for table, value in table_answer.items():
        if table not in 'unique':
            final_answer *= value
    return final_answer


def flatten_list_of_lists(some_list, remove_duplicates=False, sort=False):
    """
    Convert a list of lists into a list of all values
    :param some_list: a list such that each value is a list
    :type some_list: list
    :param remove_duplicates: if True, return a unique list, otherwise keep duplicated values
    :type remove_duplicates: bool
    :param sort: if True, sort the list
    :type sort: bool
    :return: a new object containing all values in teh provided
    """
    data = [item for sublist in some_list for item in sublist]
    if sort:
        data.sort()
    if remove_duplicates:
        return list(set(data))
    return data


def collapse_into_single_dates(x, y):
    """
    Function used for a time plot to convert multiple values into one value, while retaining enough information
    to perform a moving average over time
    :param x: a list of dates in ascending order
    :param y: a list of values and can use the '+' operator as a function of date
    :return: a unique list of dates, sum of y for that date, and number of original points for that date
    :rtype: dict
    """

    # average daily data and keep track of points per day
    x_collapsed = [x[0]]
    y_collapsed = [y[0]]
    w_collapsed = [1]
    for n in range(1, len(x)):
        if x[n] == x_collapsed[-1]:
            y_collapsed[-1] = (y_collapsed[-1] + y[n])
            w_collapsed[-1] += 1
        else:
            x_collapsed.append(x[n])
            y_collapsed.append(y[n])
            w_collapsed.append(1)

    return {'x': x_collapsed, 'y': y_collapsed, 'w': w_collapsed}


def moving_avg(xyw, avg_len):
    """
    Calculate a moving average for a given averaging length
    :param xyw: output from collapse_into_single_dates
    :type xyw: dict
    :param avg_len: average of these number of points, i.e., look-back window
    :type avg_len: int
    :return: list of x values, list of y values
    :rtype: tuple
    """
    cumsum, moving_aves, x_final = [0], [], []

    for i, y in enumerate(xyw['y'], 1):
        cumsum.append(cumsum[i - 1] + y / xyw['w'][i - 1])
        if i >= avg_len:
            moving_ave = (cumsum[i] - cumsum[i - avg_len]) / avg_len
            moving_aves.append(moving_ave)
    x_final = [xyw['x'][i] for i in range(avg_len - 1, len(xyw['x']))]

    return x_final, moving_aves


def convert_value_to_str(value, rounding_digits=2):
    try:
        formatter = "%%0.%df" % rounding_digits
        return formatter % value
    except TypeError:
        return value


def get_selected_listctrl_items(list_control):
    """
    Get the indices of the currently selected items of a wx.ListCtrl object
    :param list_control: any wx.ListCtrl object
    :type list_control: ListCtrl
    :return: indices of selected items
    :rtype: list
    """
    selection = []

    index_current = -1
    while True:
        index_next = list_control.GetNextItem(index_current, wx.LIST_NEXT_ALL, wx.LIST_STATE_SELECTED)
        if index_next == -1:
            return selection

        selection.append(index_next)
        index_current = index_next


def print_run_time(start_time, end_time, calc_title):
    """
    :param start_time: start time of process
    :type start_time: datetime
    :param end_time: end time of process
    :type end_time: datetime
    :param calc_title: prepend the status message with this value
    :type calc_title: str
    :return:
    """
    total_time = end_time - start_time
    seconds = total_time.seconds
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    if h:
        print("%s. This took %dhrs %02dmin %02dsec to complete" % (calc_title, h, m, s))
    elif m:
        print("%s. This took %02dmin %02dsec to complete" % (calc_title, m, s))
    else:
        print("%s. This took %02dsec to complete" % (calc_title, s))


def datetime_to_date_string(datetime_obj):
    if isinstance(datetime_obj, str):
        datetime_obj = parse_date(datetime_obj)
    return "%s/%s/%s" % (datetime_obj.month, datetime_obj.day, datetime_obj.year)


def change_angle_origin(angles, max_positive_angle):
    """
    Angles in DICOM are all positive values, but there is typically no mechanical continuity in across 180 degrees
    :param angles: angles to be converted
    :type angles list
    :param max_positive_angle: the maximum positive angle, angles greater than this will be shifted to negative angles
    :return: list of the same angles, but none exceed the max
    :rtype: list
    """
    if len(angles) == 1:
        if angles[0] > max_positive_angle:
            return [angles[0] - 360]
        else:
            return angles
    new_angles = []
    for angle in angles:
        if angle > max_positive_angle:
            new_angles.append(angle - 360)
        elif angle == max_positive_angle:
            if angle == angles[0] and angles[1] > max_positive_angle:
                new_angles.append(angle - 360)
            elif angle == angles[-1] and angles[-2] > max_positive_angle:
                new_angles.append(angle - 360)
            else:
                new_angles.append(angle)
        else:
            new_angles.append(angle)
    return new_angles


def calc_stats(data):
    """
    Calculate a standard set of stats for DVHA
    :param data: a list or numpy 1D array of numbers
    :type data: list
    :return:  max, 75%, median, mean, 25%, and min of data
    :rtype: list
    """
    data = [x for x in data if x != 'None']
    try:
        data_np = np.array(data)
        rtn_data = [np.max(data_np),
                    np.percentile(data_np, 75),
                    np.median(data_np),
                    np.mean(data_np),
                    np.percentile(data_np, 25),
                    np.min(data_np)]
    except Exception:
        rtn_data = [0, 0, 0, 0, 0, 0]
        print("calc_stats() received non-numerical data")
    return rtn_data


def move_files_to_new_path(files, new_dir, copy_files=False):
    """
    Move all files provided to the new directory
    :param files: absolute file paths
    :type files: list
    :param new_dir: absolute directory path
    :type new_dir: str
    :param copy_files: Set to True to keep original files and copy to new_dir, False to remove original files
    :type copy_files: bool
    """
    for file_path in files:
        if isfile(file_path):
            file_name = basename(file_path)
            old_dir = dirname(file_path)
            new = join(new_dir, file_name)
            if not isdir(new_dir):
                mkdir(new_dir)
            if old_dir != new_dir:
                [shutil.move, shutil.copy][copy_files](file_path, new)


def delete_directory_contents(dir_to_delete):
    # https://stackoverflow.com/questions/185936/how-to-delete-the-contents-of-a-folder-in-python
    for the_file in listdir(dir_to_delete):
        delete_file(join(dir_to_delete, the_file))


def delete_file(file_path):
    try:
        if isfile(file_path):
            unlink(file_path)
        elif isdir(file_path):
            shutil.rmtree(file_path)
    except Exception as e:
        print(e)


def delete_imported_dicom_files(dicom_files):
    """
    delete imported dicom files
    :param dicom_files: the return from DVH_SQL().get_dicom_file_paths
    :type dicom_files: dict
    """
    for i, directory in enumerate(dicom_files['folder_path']):
        # Delete associated plan, structure, and dose files
        for key in ['plan_file', 'structure_file', 'dose_file']:
            delete_file(join(directory, dicom_files[key][i]))

        # Delete misc dicom files for given study instance uid
        remaining_files = listdir(directory)
        for f in remaining_files:
            try:
                uid = str(pydicom.read_file(join(directory, f)).StudyInstanceUID)
                if uid == str(dicom_files['study_instance_uid'][i]):
                    delete_file(f)
            except Exception:
                pass

        # Directory is empty, delete it
        # Directories are by patient mrn, so it might contain files for a different study for the same patient
        if not listdir(directory):
            try:
                rmdir(directory)
            except Exception:
                pass


def move_imported_dicom_files(dicom_files, new_dir):
    """
    move imported dicom files
    :param dicom_files: the return from DVH_SQL().get_dicom_file_paths
    :type dicom_files: dict
    """
    for i, directory in enumerate(dicom_files['folder_path']):
        files = [join(directory, dicom_files[key][i]) for key in ['plan_file', 'structure_file', 'dose_file']]
        new_patient_dir = join(new_dir, dicom_files['mrn'][i])
        move_files_to_new_path(files, new_patient_dir)

        # Move misc dicom files for given study instance uid
        remaining_files = listdir(directory)
        files = []
        for f in remaining_files:
            try:
                uid = str(pydicom.read_file(join(directory, f)).StudyInstanceUID)
                if uid == str(dicom_files['study_instance_uid'][i]):
                    files.append(f)
            except Exception:
                pass
        move_files_to_new_path(files, new_patient_dir)

        # Directory is empty, delete it
        # Directories are by patient mrn, so it might contain files for a different study for the same patient
        if not listdir(directory):
            try:
                rmdir(directory)
            except Exception:
                pass


def remove_empty_sub_folders(start_path):
    for (path, dirs, files) in walk(start_path, topdown=False):
        if files:
            continue
        try:
            if path != start_path:
                rmdir(path)
        except OSError:
            pass


def move_all_files(new_dir, old_dir):
    """
    This function will move all files from the old to new directory, it will ignore all files in subdirectories
    :param new_dir: absolute directory path
    :param old_dir: absolute directory path
    """
    initial_path = dirname(realpath(__file__))

    chdir(old_dir)

    file_paths = [f for f in listdir(old_dir) if isfile(join(old_dir, f))]

    misc_path = join(new_dir, 'misc')
    if not isdir(misc_path):
        mkdir(misc_path)

    for f in file_paths:
        file_name = basename(f)
        new = join(misc_path, file_name)
        shutil.move(f, new)

    chdir(initial_path)


def get_elapsed_time(start_time, end_time):
    total_time = end_time - start_time
    seconds = total_time.seconds
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    if h:
        return "%d hrs %d min %d sec" % (h, m, s)
    if m:
        return "%d min %d sec" % (m, s)
    return "%d sec" % s


def is_date(date):
    if isinstance(date, datetime):
        return True

    if isinstance(date, str):
        try:
            parse_date(date)
            return True
        except Exception:
            return False

    return False


def rank_ptvs_by_D95(ptvs):
    """
    Determine the order of provided PTVs by their D_95% values
    :param ptvs: dvh, volume, index of PTVs
    :type ptvs: dict
    :return: ptv numbers in order of D_95%
    """
    doses_to_rank = get_dose_to_volume(ptvs['dvh'], ptvs['volume'], 0.95)
    return sorted(range(len(ptvs['dvh'])), key=lambda k: doses_to_rank[k])


def get_dose_to_volume(dvhs, volumes, roi_fraction):
    # Not precise (i.e., no interpolation) but good enough for sorting PTVs
    doses = []
    for i, dvh in enumerate(dvhs):
        abs_volume = volumes[i] * roi_fraction
        dvh_np = np.array(dvh.split(','), dtype=np.float)
        try:
            dose = next(x[0] for x in enumerate(dvh_np) if x[1] < abs_volume)
        except StopIteration:
            dose = dvh_np[-1]
        doses.append(dose)

    return doses


def float_or_none(value):
    try:
        return float(value)
    except ValueError:
        return 'None'


class MessageDialog:
    """
    This is the base class for Yes/No Dialog boxes
    Inherit this class, then over-write action_yes and action_no functions with appropriate behaviors
    """
    def __init__(self, parent, caption, message="Are you sure?", action_yes_func=None, action_no_func=None,
                 flags=wx.ICON_WARNING | wx.YES | wx.NO | wx.NO_DEFAULT):
        if is_windows():
            message = '\n'.join([caption, message])
            caption = ' '
        self.dlg = wx.MessageDialog(parent, message, caption, flags)
        self.parent = parent
        self.action_yes_func = action_yes_func
        self.action_no_func = action_no_func
        self.run()

    def run(self):
        res = self.dlg.ShowModal()
        [self.action_no, self.action_yes][res == wx.ID_YES]()
        self.dlg.Destroy()

    def action_yes(self):
        if self.action_yes_func is not None:
            self.action_yes_func()

    def action_no(self):
        if self.action_no_func is not None:
            self.action_no_func()


def save_object_to_file(obj, abs_file_path):
    """
    Save a python object acceptable for pickle to the provided file path
    """
    with open(abs_file_path, 'wb') as outfile:
        pickle.dump(obj, outfile)


def load_object_from_file(abs_file_path):
    """
    Load a pickled object from the provided absolute file path
    """
    if isfile(abs_file_path):
        with open(abs_file_path, 'rb') as infile:
            obj = pickle.load(infile)
        return obj


def sample_list(some_list, max_size, n):
    """
    Reduce a list by given factor iteratively until list size less than max_size
    :param some_list: any list you like!
    :type some_list: list
    :param max_size: the maximum number of items in the returned list
    :type max_size: int
    :param n: remove every nth element
    :type n: int
    :return: sampled list
    :rtype: list
    """
    while len(some_list) > max_size:
        some_list = remove_every_nth_element(some_list, n)
    return some_list


def remove_every_nth_element(some_list, n):
    return [value for i, value in enumerate(some_list) if i % n != 0]


def sample_roi(roi_coord, max_point_count=5000, iterative_reduction=0.1):
    """
    Iteratively sample a list of 3D points by the iterative_reduction until the size of the list is < max_point_count
    This is used to reduce the number of points used in the ptv distance calculations because:
        1) Shapely returns a much large number of points when calculating total PTVs
        2) Users could easily run into memory issues using scip.dist if all points are used (particularly on MSW)
    :param roi_coord: a list of 3D points representing an roi
    :type roi_coord: list
    :param max_point_count: the maximum number of points in the returned roi_coord
    :type max_point_count: int
    :param iterative_reduction: iteratively remove this fraction of points until len < max_point_count
    :type iterative_reduction: float
    :return: sampled roi
    :rtype: list
    """
    return sample_list(roi_coord, max_point_count, int(1 / iterative_reduction))


def get_sorted_indices(some_list):
    try:
        return [i[0] for i in sorted(enumerate(some_list), key=lambda x: x[1])]
    except TypeError:  # can't sort if a mix of str and float
        try:
            temp_data = [[value, -float('inf')][value == 'None'] for value in some_list]
            return [i[0] for i in sorted(enumerate(temp_data), key=lambda x: x[1])]
        except TypeError:
            temp_data = [str(value) for value in some_list]
            return [i[0] for i in sorted(enumerate(temp_data), key=lambda x: x[1])]


def get_window_size(width, height):
    """
    Function used to adapt frames/windows for the user's resolution
    :param width: fractional width of the user's screen
    :param height: fractional height of the user's screen
    :return: window size
    :rtype: tuple
    """
    user_width, user_height = wx.GetDisplaySize()
    if user_width / user_height < 1.5:  # catch 4:3 or non-widescreen
        user_height = user_width / 1.6
    return tuple([int(width * user_width), int(height * user_height)])


def set_frame_icon(frame):
    if not is_mac():
        frame.SetIcon(wx.Icon(WIN_APP_ICON))


def trace_memory_alloc_pretty_top(snapshot, key_type='lineno', limit=10):
    """From https://docs.python.org/3/library/tracemalloc.html"""
    print('-----------------------------------------------------------------')
    snapshot = snapshot.filter_traces((
        tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
        tracemalloc.Filter(False, "<unknown>"),
    ))
    top_stats = snapshot.statistics(key_type)

    print("Top %s lines" % limit)
    for index, stat in enumerate(top_stats[:limit], 1):
        frame = stat.traceback[0]
        # replace "/path/to/module/file.py" with "module/file.py"
        filename = sep.join(frame.filename.split(sep)[-2:])
        print("#%s: %s:%s: %.1f KiB"
              % (index, filename, frame.lineno, stat.size / 1024))
        line = linecache.getline(frame.filename, frame.lineno).strip()
        if line:
            print('    %s' % line)

    other = top_stats[limit:]
    if other:
        size = sum(stat.size for stat in other)
        print("%s other: %.1f KiB" % (len(other), size / 1024))
    total = sum(stat.size for stat in top_stats)
    print("Total allocated size: %.1f KiB" % (total / 1024))


def trace_memory_alloc_simple_stats(snapshot, key_type='lineno'):
    """From https://docs.python.org/3/library/tracemalloc.html"""
    print('-----------------------------------------------------------------')
    print("[ Top 10 ]")
    top_stats = snapshot.statistics(key_type)
    for stat in top_stats[:10]:
        print(stat)


class PopupMenu:
    def __init__(self, parent):
        self.parent = parent
        self.menus = []

    def add_menu_item(self, label, action):
        self.menus.append({'id': wx.NewId(), 'label': label, 'action': action})

    def run(self):
        popup_menu = wx.Menu()
        for menu in self.menus:
            popup_menu.Append(menu['id'], menu['label'])
            self.parent.Bind(wx.EVT_MENU, menu['action'], id=menu['id'])

        self.parent.PopupMenu(popup_menu)
        popup_menu.Destroy()


def validate_transfer_syntax_uid(data_set):
    meta = pydicom.Dataset()
    meta.ImplementationClassUID = pydicom.uid.generate_uid()
    meta.TransferSyntaxUID = ImplicitVRLittleEndian
    new_data_set = pydicom.FileDataset(filename_or_obj=None, dataset=data_set, is_little_endian=True,
                                       file_meta=meta)
    new_data_set.is_little_endian = True
    new_data_set.is_implicit_VR = True

    return new_data_set


def get_installed_python_libraries():
    """Use pip command line function 'list' to extract the currently installed libraries"""

    try:
        output = str(check_output(['pip', 'list', '--local']), 'utf-8').split('\n')
    except Exception:
        # If running from PyInstaller, this will fail, pickle a file prior to freezing with save_pip_list
        return load_object_from_file(PIP_LIST_PATH)

    python_version = '.'.join(str(i) for i in sys.version_info[:3])
    libraries = {'Library': ['python'], 'Version': [python_version]}
    for row in output[2:]:  # ignore first two rows which are column headers and a separator
        data = [v for v in row.strip().split(' ') if v]
        if data:
            libraries['Library'].append(data[0])
            libraries['Version'].append(data[1])
    return libraries


def save_pip_list():
    save_object_to_file(get_installed_python_libraries(), PIP_LIST_PATH)



from os.path import join
from dvha.paths import TG263_CSV, PREF_DIR


class ROIMapGenerator:
    """Class to interact with the TG263 table and generate compliant ROI Name Maps"""
    def __init__(self):
        # Load TG263 table
        with open(TG263_CSV, 'r') as doc:
            keys = [key.strip() for key in doc.readline().split(',')]
            self.tg_263 = {key: [] for key in keys}
            for line in doc:
                if 'xx' not in line:  # ignore the rows with generic expansions
                    for col, value in enumerate(line.split(',')):
                        if value == '':
                            value = 'None'
                        self.tg_263[keys[col]].append(value.strip().replace('^', ','))
            self.keys = keys
            self.keys.append(self.keys.pop(0))  # Move Target Type to the end

        self.key_map = {key: key for key in keys}

    def __call__(self, map_file_name, body_sites=None, data_filter=None, roi_uid_type='primary'):
        """
        Create a new ROI map file based on TG263
        :param map_file_name: save the file in PREF_DIR with this name
        :type map_file_name: str
        :param body_sites: automatically create a data_filter based on these body sites
        :type body_sites: list
        :param data_filter: used to write a subset of the TG263 data
        :type data_filter: dict
        :param roi_uid_type: either 'primary', 'reverse', or 'fmaid'
        :type roi_uid_type: str
        :return: file_path of new map file
        """
        if body_sites is not None:
            if type(body_sites) is not list:
                body_sites = [body_sites]
            data_filter = {'Anatomic Group': body_sites}
        data = self.tg_263 if data_filter is None else self.get_filtered_data(data_filter)

        lookup = {'primary': 'TG263-Primary Name',
                  'reverse': 'TG-263-Reverse Order Name',
                  'fmaid': 'FMAID'}
        roi_uids = self.get_unique_values(lookup[roi_uid_type], data)

        file_path = join(PREF_DIR, map_file_name)
        with open(file_path, 'w') as doc:
            for roi_uid in roi_uids:
                doc.write(": ".join([roi_uid] * 3) + '\n')
        return file_path

    def prep_data_for_roi_map_gui(self):
        ignored = [c for c in self.keys if 'Reverse' in c or 'Character' in c or 'FMAID' in c]
        for c in ignored:
            self.drop_column(c)

        str_map = {'Category': 'Cat.', 'Anatomic': 'Anat.', 'TG263-': ''}
        for c in self.keys:
            if any([key in c for key in list(str_map)]):
                new_key = c
                for key, value in str_map.items():
                    new_key = new_key.replace(key, value)
                self.key_map[c] = new_key
                key_index = self.keys.index(c)
                self.tg_263[new_key] = self.drop_column(c, return_data=True)
                self.keys.insert(key_index, new_key)

    ##########################################################
    # Generalized Tools
    ##########################################################
    def get_filtered_data(self, data_filter):
        """
        Get TG263 data with a filter
        :param data_filter: {column: list_allowed_values}
        :type data_filter: dict
        :return: subset of tg_263 with the data_filter applied
        :type: dict
        """

        for key in list(data_filter):
            if data_filter[key].lower() == 'all':
                data_filter.pop(key)

        data = {key: [] for key in self.keys}
        for row in range(len(self.tg_263[self.keys[0]])):
            is_included = [self.tg_263[col][row] in data_filter[col] for col in list(data_filter)]
            if all(is_included):
                for c in self.keys:
                    data[c].append(self.tg_263[c][row])
        return data

    def get_unique_values(self, column, data=None):
        """
        :param column: TG263 column
        :param data: optionally provide filtered data, default is entire TG263 table
        :return: list of unique values for the provided column
        """
        data = self.tg_263 if data is None else data
        return sorted([value for value in set(data[column]) if value])

    def get_value_from_uid(self, roi_uid, output_type, reverse_name=False):
        """
        :param roi_uid: either a primary name or FMAID (auto-detects)
        :type roi_uid: str
        :param output_type: column name of output value
        :type output_type: str
        :param reverse_name: roi_uid is assumed to be reverse order name if True
        :type reverse_name: bool
        :return: another column value of the provided roi_uid
        """
        input_type = 'FMAID' if roi_uid.isdigit() else ['TG-263-Reverse Order Name', 'TG263-Primary Name'][reverse_name]
        return self._get_value_from_uid(roi_uid, input_type=self.key_map[input_type],
                                        output_type=self.key_map[output_type])

    def _get_value_from_uid(self, input_value, input_type, output_type):
        """Generic function to look up another column with a given input value and type"""
        if input_value in self.tg_263[input_type]:
            index = self.tg_263[input_type].index(input_value)
            return self.tg_263[output_type][index]

    def drop_column(self, column, return_data=False):
        """Remove a column from tg_263 data and update keys"""
        if column in self.keys and column in list(self.tg_263):
            self.keys.pop(self.keys.index(column))
            popped_data = self.tg_263.pop(column)
        if return_data:
            return popped_data

    ##########################################################
    # Properties for coding ease
    ##########################################################
    @property
    def target_types(self):
        return self.get_unique_values(self.key_map['Target Type'])

    @property
    def anatomic_groups(self):
        return self.get_unique_values(self.key_map['Anatomic Group'])

    @property
    def major_categories(self):
        return self.get_unique_values(self.key_map['Major Category'])

    @property
    def minor_categories(self):
        return self.get_unique_values(self.key_map['Minor Category'])

    @property
    def primary_names(self):
        return self.get_unique_values(self.key_map['TG263-Primary Name'])

    @property
    def reverse_order_primary_names(self):
        return self.get_unique_values(self.key_map['TG-263-Reverse Order Name'])

    @property
    def fmaids(self):
        return self.get_unique_values(self.key_map['FMAID'])

    def get_primary_name(self, fmaid):
        return self._get_value_from_uid(fmaid, input_type='FMAID', output_type='TG263-Primary Name')

    def get_fmaid(self, primary_name):
        return self._get_value_from_uid(primary_name, input_type='TG263-Primary Name', output_type='FMAID')

    def get_target_type(self, roi_uid):
        return self.get_value_from_uid(roi_uid, output_type='Target Type')

    def get_major_category(self, roi_uid):
        return self.get_value_from_uid(roi_uid, output_type='Major Category')

    def get_minor_category(self, roi_uid):
        return self.get_value_from_uid(roi_uid, output_type='Minor Category')

    def get_anatomic_group(self, roi_uid):
        return self.get_value_from_uid(roi_uid, output_type='Anatomic Group')

    def get_reverse_order_name(self, roi_uid):
        return self.get_value_from_uid(roi_uid, output_type='TG-263-Reverse Order Name')

    def get_description(self, roi_uid):
        return self.get_value_from_uid(roi_uid, output_type='Description')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# models.regression.py
"""
Class to view and calculate linear regressions
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import wx
from functools import partial
from pubsub import pub
from dvha.tools.errors import ErrorDialog
from dvha.models.plot import PlotRegression, PlotMultiVarRegression
from dvha.models.machine_learning import RandomForestFrame, GradientBoostingFrame, DecisionTreeFrame,\
    SupportVectorRegressionFrame, MachineLearningModelViewer
from dvha.dialogs.export import save_data_to_file
from dvha.dialogs.main import SelectRegressionVariablesDialog
from dvha.options import DefaultOptions
from dvha.paths import ICONS, MODELS_DIR
from dvha.tools.stats import MultiVariableRegression
from dvha.tools.utilities import set_msw_background_color, get_tree_ctrl_image, get_window_size, load_object_from_file,\
    set_frame_icon


class RegressionFrame:
    """
    Object to be passed into notebook panel for the Regression tab
    """
    def __init__(self, parent, group_data, options, main_app_frame):
        """
        :param parent:  notebook panel in main view
        :type parent: Panel
        :param group_data: object containing queried data applicable/parsed for statistical analysis
        :type group_data: dict
        :param options: user options containing visual preferences
        :type options: Options
        """
        self.parent = parent
        self.options = options
        self.group_data = group_data
        self.main_app_frame = main_app_frame
        self.group = 1
        self.choices = []

        self.y_variable_nodes = {}
        self.x_variable_nodes = {}

        self.__define_gui_objects()
        self.__set_properties()
        self.__do_bind()
        self.__do_layout()

        self.tree_ctrl_root = self.tree_ctrl.AddRoot('Regressions')

        self.mvr_frames = []

        set_msw_background_color(self.window, color='white')

    def __define_gui_objects(self):
        self.window = wx.SplitterWindow(self.parent, wx.ID_ANY)
        self.pane_tree = wx.ScrolledWindow(self.window, wx.ID_ANY, style=wx.TAB_TRAVERSAL)
        self.tree_ctrl = wx.TreeCtrl(self.pane_tree, wx.ID_ANY)
        self.pane_plot = wx.Panel(self.window, wx.ID_ANY)
        self.combo_box_x_axis = wx.ComboBox(self.pane_plot, wx.ID_ANY, choices=[],
                                            style=wx.CB_DROPDOWN | wx.TE_READONLY)
        self.spin_button_x_axis = wx.SpinButton(self.pane_plot, wx.ID_ANY, style=wx.SP_WRAP)
        self.combo_box_y_axis = wx.ComboBox(self.pane_plot, wx.ID_ANY, choices=[],
                                            style=wx.CB_DROPDOWN | wx.TE_READONLY)
        self.spin_button_y_axis = wx.SpinButton(self.pane_plot, wx.ID_ANY, style=wx.SP_WRAP)
        self.checkbox = wx.CheckBox(self.pane_plot, wx.ID_ANY, "Include", style=wx.ALIGN_RIGHT)
        self.plot = PlotRegression(self.pane_plot, self.options)
        self.button_multi_var_reg_model = wx.Button(self.pane_tree, wx.ID_ANY, 'Run Multi-Variable Regressions')
        self.button_multi_var_quick_select = wx.Button(self.pane_tree, wx.ID_ANY, 'Variable Quick Select')
        self.button_single_var_export = wx.Button(self.pane_tree, wx.ID_ANY, 'Export Plot Data')
        self.button_single_var_plot_save = wx.Button(self.pane_tree, wx.ID_ANY, 'Save Plot')

    def __set_properties(self):
        self.pane_tree.SetScrollRate(20, 20)
        self.window.SetMinimumPaneSize(20)
        self.combo_box_x_axis.SetValue('ROI Volume')
        self.combo_box_y_axis.SetValue('ROI Max Dose')

        self.image_list = wx.ImageList(16, 16)
        self.images = {'y': self.image_list.Add(get_tree_ctrl_image(ICONS['custom_Y'])),
                       'x': self.image_list.Add(get_tree_ctrl_image(ICONS['custom_X']))}
        self.tree_ctrl.AssignImageList(self.image_list)

    def __do_bind(self):
        self.parent.Bind(wx.EVT_COMBOBOX, self.on_combo_box, id=self.combo_box_x_axis.GetId())
        self.parent.Bind(wx.EVT_COMBOBOX, self.on_combo_box, id=self.combo_box_y_axis.GetId())
        self.parent.Bind(wx.EVT_SPIN, self.spin_x, id=self.spin_button_x_axis.GetId())
        self.parent.Bind(wx.EVT_SPIN, self.spin_y, id=self.spin_button_y_axis.GetId())
        self.parent.Bind(wx.EVT_CHECKBOX, self.on_checkbox, id=self.checkbox.GetId())
        self.pane_tree.Bind(wx.EVT_BUTTON, self.on_regression, id=self.button_multi_var_reg_model.GetId())
        self.pane_tree.Bind(wx.EVT_BUTTON, self.on_quick_select, id=self.button_multi_var_quick_select.GetId())
        self.pane_tree.Bind(wx.EVT_TREE_SEL_CHANGED, self.on_tree_select, id=self.tree_ctrl.GetId())

        self.pane_tree.Bind(wx.EVT_BUTTON, self.on_export, id=self.button_single_var_export.GetId())
        self.pane_tree.Bind(wx.EVT_BUTTON, self.on_save_plot, id=self.button_single_var_plot_save.GetId())

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_plot = wx.BoxSizer(wx.HORIZONTAL)
        sizer_plot_pane = wx.BoxSizer(wx.VERTICAL)
        sizer_plot_view = wx.BoxSizer(wx.HORIZONTAL)
        sizer_input = wx.BoxSizer(wx.HORIZONTAL)
        sizer_check_box = wx.BoxSizer(wx.HORIZONTAL)

        sizer_y_axis = wx.BoxSizer(wx.VERTICAL)
        sizer_y_axis_select = wx.BoxSizer(wx.HORIZONTAL)
        sizer_x_axis = wx.BoxSizer(wx.VERTICAL)
        sizer_x_axis_select = wx.BoxSizer(wx.HORIZONTAL)

        sizer_single_var_export = wx.BoxSizer(wx.HORIZONTAL)

        sizer_tree = wx.BoxSizer(wx.VERTICAL)
        sizer_tree.Add(self.button_multi_var_reg_model, 0, wx.EXPAND | wx.TOP | wx.LEFT | wx.RIGHT, 5)
        sizer_tree.Add(self.button_multi_var_quick_select, 0, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.BOTTOM, 5)
        sizer_tree.Add(self.tree_ctrl, 1, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
        sizer_single_var_export.Add(self.button_single_var_export, 0, wx.EXPAND | wx.BOTTOM | wx.RIGHT, 5)
        sizer_single_var_export.Add(self.button_single_var_plot_save, 1, wx.EXPAND | wx.LEFT | wx.BOTTOM, 5)
        sizer_tree.Add(sizer_single_var_export, 0, wx.EXPAND | wx.ALL, 5)
        self.pane_tree.SetSizer(sizer_tree)

        label_x_axis = wx.StaticText(self.pane_plot, wx.ID_ANY, "Independent Variable (x-axis):")
        sizer_check_box.Add(label_x_axis, 1, wx.EXPAND, 0)
        sizer_check_box.Add(self.checkbox, 0, wx.LEFT | wx.RIGHT, 10)
        sizer_x_axis.Add(sizer_check_box, 1, wx.EXPAND, 0)
        sizer_x_axis_select.Add(self.combo_box_x_axis, 1, wx.EXPAND, 0)
        sizer_x_axis_select.Add(self.spin_button_x_axis, 0, wx.EXPAND, 0)
        sizer_x_axis.Add(sizer_x_axis_select, 1, wx.EXPAND, 0)
        sizer_input.Add(sizer_x_axis, 1, wx.ALL, 5)

        sizer_input.Add((30, 10), 0, 0, 0)

        label_y_axis = wx.StaticText(self.pane_plot, wx.ID_ANY, "Dependent Variable (y-axis):")
        sizer_y_axis.Add(label_y_axis, 1, 0, 0)
        sizer_y_axis_select.Add(self.combo_box_y_axis, 1, wx.EXPAND, 0)
        sizer_y_axis_select.Add(self.spin_button_y_axis, 0, wx.EXPAND, 0)
        sizer_y_axis.Add(sizer_y_axis_select, 1, wx.EXPAND, 0)
        sizer_input.Add(sizer_y_axis, 1, wx.ALL, 5)

        sizer_plot_pane.Add(sizer_input, 0, wx.EXPAND, 0)
        sizer_plot.Add(self.plot.layout, 1, wx.EXPAND, 0)
        sizer_plot_view.Add(sizer_plot, 1, wx.EXPAND, 10)
        sizer_plot_pane.Add(sizer_plot_view, 1, wx.EXPAND, 0)
        self.pane_plot.SetSizer(sizer_plot_pane)
        self.window.SplitVertically(self.pane_tree, self.pane_plot)
        sizer_wrapper.Add(self.window, 1, wx.EXPAND, 0)
        self.window.SetSashPosition(250)

        self.layout = sizer_wrapper

    @property
    def x_axis(self):
        return self.combo_box_x_axis.GetValue()

    @property
    def y_axis(self):
        return self.combo_box_y_axis.GetValue()

    def update_combo_box_choices(self):
        if self.group_data[1]['stats_data']:
            self.choices = self.group_data[1]['stats_data'].variables
            self.choices.sort()
            self.combo_box_x_axis.SetItems(self.choices)
            self.combo_box_y_axis.SetItems(self.choices)
            self.spin_button_x_axis.SetMax(len(self.choices)-1)
            self.spin_button_y_axis.SetMax(len(self.choices)-1)
            if 'ROI Volume' in self.choices and 'ROI Max Dose' in self.choices:
                initial_index_x = self.choices.index('ROI Volume')
                initial_index_y = self.choices.index('ROI Max Dose')
                self.spin_button_x_axis.SetValue(len(self.choices)-1-initial_index_x)
                self.spin_button_y_axis.SetValue(len(self.choices)-1-initial_index_y)
                self.combo_box_x_axis.SetValue(self.choices[initial_index_x])
                self.combo_box_y_axis.SetValue(self.choices[initial_index_y])
            self.update_plot()

    def on_combo_box(self, evt):
        self.sync_spin_buttons()
        self.update_plot()

    def update_plot(self):
        if self.combo_box_x_axis.GetValue() == self.combo_box_y_axis.GetValue():
            self.plot.clear_plot()
        else:
            stats_data = {grp: self.group_data[grp]['stats_data'] for grp in [1, 2]}
            plot_data = {1: stats_data[1].get_bokeh_data(self.x_axis, self.y_axis)}
            if stats_data[2] is not None:
                plot_data[2] = stats_data[2].get_bokeh_data(self.x_axis, self.y_axis)
            else:
                plot_data[2] = None
            self.plot.update_plot(plot_data, self.group,
                                  self.combo_box_x_axis.GetValue(),
                                  stats_data[1].get_axis_title(self.x_axis),
                                  stats_data[1].get_axis_title(self.y_axis))

        if self.y_axis in list(self.y_variable_nodes) and self.x_axis in list(self.x_variable_nodes[self.y_axis]):
            self.checkbox.SetValue(True)
            self.tree_ctrl.SelectItem(self.x_variable_nodes[self.y_axis][self.x_axis])
        else:
            self.checkbox.SetValue(False)
            self.tree_ctrl.Unselect()
        if self.combo_box_x_axis.GetValue() == self.combo_box_y_axis.GetValue():
            self.checkbox.Disable()
        else:
            self.checkbox.Enable()

    def spin_x(self, evt):
        new_index = len(self.choices)-1 - int(self.spin_button_x_axis.GetValue())
        self.combo_box_x_axis.SetValue(self.choices[new_index])
        self.update_plot()

    def spin_y(self, evt):
        new_index = len(self.choices)-1 - int(self.spin_button_y_axis.GetValue())
        self.combo_box_y_axis.SetValue(self.choices[new_index])
        self.update_plot()

    def sync_spin_buttons(self):
        if self.x_axis:
            index = self.choices.index(self.x_axis)
            self.spin_button_x_axis.SetValue(len(self.choices)-1 - index)

            index = self.choices.index(self.y_axis)
            self.spin_button_y_axis.SetValue(len(self.choices)-1 - index)
        else:
            print('ERROR: x-axis choice is empty.')

    def on_checkbox(self, *evt):
        y_value = self.combo_box_y_axis.GetValue()
        x_value = self.combo_box_x_axis.GetValue()
        [self.del_regression, self.add_regression][self.checkbox.GetValue()](y_value, x_value)

    def add_regression(self, y_var, x_var, select_item=True):
        """
        Add the currently plotted variables to the TreeCtrl to be available for multi-variable regression
        :param y_var: dependent variable
        :type y_var: str
        :param x_var: independent variable
        :type x_var: str
        :param select_item: If True, select the item in the TreeCtrl
        """

        if y_var not in list(self.y_variable_nodes):
            self.y_variable_nodes[y_var] = self.tree_ctrl.AppendItem(self.tree_ctrl_root, y_var)
            self.tree_ctrl.SetItemData(self.y_variable_nodes[y_var], None)
            self.tree_ctrl.SetItemImage(self.y_variable_nodes[y_var], self.images['y'], wx.TreeItemIcon_Normal)
        if y_var not in list(self.x_variable_nodes):
            self.x_variable_nodes[y_var] = {}
        if x_var not in self.x_variable_nodes[y_var]:
            self.x_variable_nodes[y_var][x_var] = self.tree_ctrl.AppendItem(self.y_variable_nodes[y_var], x_var)
            self.tree_ctrl.SetItemData(self.x_variable_nodes[y_var][x_var], None)
            self.tree_ctrl.SetItemImage(self.x_variable_nodes[y_var][x_var], self.images['x'],
                                        wx.TreeItemIcon_Normal)
        self.tree_ctrl.ExpandAll()
        if select_item and self.tree_ctrl.IsSelected(self.tree_ctrl_root):
            self.tree_ctrl.SelectItem(self.x_variable_nodes[self.y_axis][self.x_axis])

    def del_regression(self, y_var, x_var):
        """
        Remove x_var from the independent variables of the y_var model, update the tree ctrl
        :param y_var: dependent variable
        :type y_var: str
        :param x_var: independent variable
        :type x_var: str
        """
        if y_var in list(self.y_variable_nodes):
            if x_var in list(self.x_variable_nodes[y_var]):
                self.tree_ctrl.Delete(self.x_variable_nodes[y_var][x_var])
                self.x_variable_nodes[y_var].pop(x_var)
                if not list(self.x_variable_nodes[y_var]):
                    self.x_variable_nodes.pop(y_var)
                    self.tree_ctrl.Delete(self.y_variable_nodes[y_var])
                    self.y_variable_nodes.pop(y_var)
        self.tree_ctrl.Unselect()

    def on_regression(self, evt):
        """
        Launch the multi-variable regression for all x_variable_nodes
        """
        if not list(self.x_variable_nodes):
            wx.MessageBox('No data has been selected for regression.', 'Regression Error',
                          wx.OK | wx.ICON_WARNING)
        else:
            for y_variable in list(self.x_variable_nodes):
                x_variables = list(self.x_variable_nodes[y_variable])

                try:
                    self.mvr_frames.append(MultiVarResultsFrame(self.main_app_frame, y_variable, x_variables,
                                                                self.group_data, self.group, self.options))
                    self.mvr_frames[-1].Show()
                except Exception as e:
                    msg = "Failed on regression for %s\n%s" % (y_variable, str(e))
                    ErrorDialog(self.parent, msg, "Multi-Variable Regression Error")

    def on_tree_select(self, evt):
        selection = evt.GetItem()
        x_var, y_var = self.get_x_y_of_node(selection)
        if y_var is not None:
            self.combo_box_y_axis.SetValue(y_var)
        if x_var is not None:
            self.combo_box_x_axis.SetValue(x_var)

        if any([x_var, y_var]):
            self.update_plot()

        self.sync_spin_buttons()

    def get_x_y_of_node(self, node):
        for y_var in list(self.x_variable_nodes):
            if node == self.y_variable_nodes[y_var]:
                return None, y_var
            for x_var, x_node in self.x_variable_nodes[y_var].items():
                if node == x_node:
                    return x_var, y_var
        return None, None

    def clear(self, new_group_data):
        self.group_data = new_group_data
        self.group = 1
        self.plot.clear_plot()
        self.x_variable_nodes = {}
        self.y_variable_nodes = {}
        self.tree_ctrl.DeleteAllItems()
        self.tree_ctrl_root = self.tree_ctrl.AddRoot('Regressions')
        self.checkbox.SetValue(False)
        self.combo_box_x_axis.SetValue('ROI Volume')
        self.combo_box_y_axis.SetValue('ROI Max Dose')

    def get_y_vars(self):
        return list(self.y_variable_nodes)

    def get_x_vars(self, y_var):
        return list(self.x_variable_nodes[y_var])

    def get_save_data(self):
        return {y_var: [x_var for x_var in self.get_x_vars(y_var)] for y_var in self.get_y_vars()}

    def load_save_data(self, save_data):
        for y_var, x_vars in save_data.items():
            for x_var in x_vars:
                self.add_regression(y_var, x_var, select_item=False)

    @property
    def has_data(self):
        return bool(len(list(self.y_variable_nodes)))

    def on_save_plot(self, evt):
        save_data_to_file(self.pane_tree, 'Save linear regression plot', self.plot.html_str,
                          wildcard="HTML files (*.html)|*.html")

    def on_export(self, evt):
        save_data_to_file(self.pane_tree, 'Export linear regression data', self.plot.get_csv_data())

    def on_quick_select(self, evt):
        if self.y_axis in list(self.x_variable_nodes):
            selections = list(self.x_variable_nodes[self.y_axis])
        else:
            selections = None
        choices = [choice for choice in self.combo_box_x_axis.GetItems() if choice != self.y_axis]
        choices.sort()
        dlg = SelectRegressionVariablesDialog(self.y_axis, choices, selections=selections)
        res = dlg.ShowModal()

        if res == wx.ID_OK:
            for value in dlg.independent_variable_choices:
                if value in dlg.selected_values:
                    if self.y_axis not in list(self.x_variable_nodes) or \
                        (self.y_axis in list(self.x_variable_nodes) and
                         value not in list(self.x_variable_nodes[self.y_axis])):
                        self.add_regression(dlg.dependent_variable, value, select_item=False)
                else:
                    if self.y_axis in list(self.x_variable_nodes) and value in list(self.x_variable_nodes[self.y_axis]):
                        self.del_regression(dlg.dependent_variable, value)
            if self.x_axis in dlg.selected_values and not self.checkbox.GetValue():
                self.checkbox.SetValue(True)
                self.on_checkbox()
        dlg.Destroy()

    def close_mvr_frames(self):
        for frame in self.mvr_frames:
            if type(frame) is MultiVarResultsFrame:
                frame.close_ml_frames()
                try:
                    frame.Close()
                except RuntimeError:
                    pass

    def apply_plot_options(self):
        self.plot.apply_options()
        for mvr_frame in self.mvr_frames:
            try:
                mvr_frame.apply_plot_options()
            except RuntimeError:
                pass


class MultiVarResultsFrame(wx.Frame):
    """
    Class to view multi-variable regression with data passed from RegressionFrame
    """
    def __init__(self, main_app_frame, y_variable, x_variables, group_data, group, options, auto_update_plot=True):
        """
        :param y_variable: dependent variable
        :type y_variable: str
        :param x_variables: independent variables
        :type x_variables: list
        :param group_data: dvhs, table, and stats_data
        :type group_data: dict
        :param options: user options containing visual preferences
        :type options: Options
        """
        wx.Frame.__init__(self, None, title="Multi-Variable Model for %s: Group %s" % (y_variable, group))

        self.main_app_frame = main_app_frame
        self.y_variable = y_variable
        self.x_variables = x_variables
        self.group_data = group_data
        self.group = group
        self.stats_data = group_data[group]['stats_data']

        set_msw_background_color(self)  # If windows, change the background color
        set_frame_icon(self)

        self.options = options

        self.plot = PlotMultiVarRegression(self, options, group)
        if auto_update_plot:
            self.plot.update_plot(y_variable, x_variables, self.stats_data)
            msg = {'y_variable': y_variable,
                   'x_variables': x_variables,
                   'regression': self.plot.reg,
                   'group': group}
            pub.sendMessage('control_chart_set_model', **msg)

        self.button_back_elimination = wx.Button(self, wx.ID_ANY, 'Backward Elimination')
        self.button_export = wx.Button(self, wx.ID_ANY, 'Export Plot Data')
        self.button_save_html = wx.Button(self, wx.ID_ANY, 'Save Plot to HTML')
        self.button_save_svg = wx.Button(self, wx.ID_ANY, 'Save Plot to SVG')
        self.button_save_model = wx.Button(self, wx.ID_ANY, 'Save MVR Model')
        self.button_load_mlr_model = wx.Button(self, wx.ID_ANY, 'Load ML Model')
        algorithms = ['Random Forest', 'Support Vector Machine', 'Decision Tree', 'Gradient Boosting']
        self.button = {key: wx.Button(self, wx.ID_ANY, key) for key in algorithms}
        self.radiobox_include_back_elim = wx.RadioBox(self, wx.ID_ANY, 'Include all x-variables?', choices=['Yes', 'No'])

        self.__do_bind()
        self.__set_properties()
        self.__do_layout()

        self.ml_frames = []

    def __set_properties(self):
        self.SetMinSize(get_window_size(0.491, 0.690))
        self.radiobox_include_back_elim.SetSelection(0)
        self.radiobox_include_back_elim.Disable()
        self.radiobox_include_back_elim.SetToolTip("Set to No if you want to the machine learning to only include "
                                                   "variables that survived the backward elimination.")

    def __do_bind(self):
        self.Bind(wx.EVT_BUTTON, self.on_random_forest, id=self.button['Random Forest'].GetId())
        self.Bind(wx.EVT_BUTTON, self.on_gradient_boosting, id=self.button['Gradient Boosting'].GetId())
        self.Bind(wx.EVT_BUTTON, self.on_decision_tree, id=self.button['Decision Tree'].GetId())
        self.Bind(wx.EVT_BUTTON, self.on_support_vector_regression,
                  id=self.button['Support Vector Machine'].GetId())
        self.Bind(wx.EVT_BUTTON, self.on_back_elimination, id=self.button_back_elimination.GetId())
        self.Bind(wx.EVT_BUTTON, self.on_export, id=self.button_export.GetId())
        self.Bind(wx.EVT_BUTTON, self.on_save_html, id=self.button_save_html.GetId())
        self.Bind(wx.EVT_BUTTON, self.on_save_svg, id=self.button_save_svg.GetId())
        self.Bind(wx.EVT_BUTTON, self.on_save_model, id=self.button_save_model.GetId())
        self.Bind(wx.EVT_BUTTON, self.on_load_mlr_model, id=self.button_load_mlr_model.GetId())
        self.Bind(wx.EVT_SIZE, self.on_resize)

    def __do_layout(self):

        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_wrapper.Add(self.plot.layout, 1, wx.EXPAND | wx.ALL, 5)

        sizer_algo_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_algo_select = wx.BoxSizer(wx.HORIZONTAL)
        sizer_export_buttons = wx.BoxSizer(wx.HORIZONTAL)
        sizer_export_buttons.Add(self.button_back_elimination, 0, wx.ALL, 5)
        sizer_export_buttons.Add(self.button_export, 0, wx.ALL, 5)
        sizer_export_buttons.Add(self.button_save_html, 0, wx.ALL, 5)
        sizer_export_buttons.Add(self.button_save_svg, 0, wx.ALL, 5)
        sizer_export_buttons.Add(self.button_save_model, 0, wx.ALL, 5)
        sizer_algo_wrapper.Add(sizer_export_buttons, 0, wx.ALL, 5)
        text = wx.StaticText(self, wx.ID_ANY, "Compare with Machine Learning Module")
        sizer_algo_wrapper.Add(text, 0, wx.EXPAND | wx.ALL, 5)
        sizer_algo_select.Add(self.button_load_mlr_model, 0, wx.EXPAND | wx.ALL, 5)
        for key, button in self.button.items():
            sizer_algo_select.Add(button, 0, wx.EXPAND | wx.ALL, 5)
        sizer_algo_select.Add(self.radiobox_include_back_elim, 0, wx.EXPAND | wx.ALL, 5)
        sizer_algo_wrapper.Add(sizer_algo_select, 0, wx.ALL, 5)

        sizer_wrapper.Add(sizer_algo_wrapper, 0, wx.EXPAND | wx.ALL, 10)

        self.SetSizer(sizer_wrapper)
        self.Fit()
        self.Layout()
        self.Center()

    @property
    def ml_include_all(self):
        return bool(1 - self.radiobox_include_back_elim.GetSelection())

    @property
    def final_stats_data(self):
        return self.plot.get_final_stats_data(include_all=self.ml_include_all)

    def on_random_forest(self, evt):
        self.ml_frames.append(RandomForestFrame(self.main_app_frame, self.final_stats_data))

    def on_gradient_boosting(self, evt):
        self.ml_frames.append(GradientBoostingFrame(self.main_app_frame, self.final_stats_data))

    def on_decision_tree(self, evt):
        self.ml_frames.append(DecisionTreeFrame(self.main_app_frame, self.final_stats_data))

    def on_support_vector_regression(self, evt):
        self.ml_frames.append(SupportVectorRegressionFrame(self.main_app_frame, self.final_stats_data))

    def on_export(self, evt):
        save_data_to_file(self, 'Save multi-variable regression data to csv', self.plot.get_csv_data())

    def on_back_elimination(self, *evt):
        wait = wx.BusyInfo('Please Wait\nPerforming Backward Elimination')
        self.plot.backward_elimination()
        del wait
        self.radiobox_include_back_elim.Enable()

    def on_save_html(self, *evt):
        try:
            if self.main_app_frame.export_figure is None:
                save_data_to_file(self, 'Save multi-variable regression plot to .html', self.plot.html_str,
                                  initial_dir="", wildcard="HTML files (*.html)|*.html")
            else:
                fig_attr_dict = self.main_app_frame.export_figure.fig_attr_dict
                func = partial(self.plot.save_figure, 'html', fig_attr_dict)
                save_data_to_file(self, 'Save multi-variable regression plot to .html', func,
                                  initial_dir="", data_type='function', wildcard="HTML files (*.html)|*.html")
        except Exception as e:
            ErrorDialog(self, str(e), "Save Error")

    def on_save_svg(self, *evt):
        try:
            if self.main_app_frame.export_figure is None:
                save_data_to_file(self, 'Save multi-variable regression plot to .svg', self.plot.export_svg,
                                  initial_dir="", data_type='function', wildcard="SVG files (*.svg)|*.svg")
            else:
                fig_attr_dict = self.main_app_frame.export_figure.fig_attr_dict
                func = partial(self.plot.save_figure, 'svg', fig_attr_dict)
                save_data_to_file(self, 'Save multi-variable regression plot to .svg', func,
                                  initial_dir="", data_type='function', wildcard="SVG files (*.svg)|*.svg")
        except Exception as e:
            ErrorDialog(self, str(e), "Save Error")

    def on_save_model(self, evt):
        data = {'y_variable': self.plot.y_variable,
                'regression': self.plot.reg,
                'x_variables': self.plot.x_variables,
                'regression_type': 'multi-variable-linear',
                'version': DefaultOptions().VERSION}
        save_data_to_file(self, 'Save Model', data,
                          wildcard="MVR files (*.mvr)|*.mvr", data_type='pickle', initial_dir=MODELS_DIR)

    def on_load_mlr_model(self, evt):
        # with wx.FileDialog(self, "Load a machine learning model", "", wildcard='*.mlr',
        #                    style=wx.FD_FILE_MUST_EXIST | wx.FD_OPEN) as dlg:
        #     dlg.SetDirectory(MODELS_DIR)
        #     if dlg.ShowModal() == wx.ID_OK:
        #         model_file_path = dlg.GetPath()
        #         saved_data = load_object_from_file(model_file_path)
        #         print(type(saved_data['regressor']))
        #         self.ml_frames.append(RandomForestFrame(self.final_stats_data))
        MachineLearningModelViewer(self, self.group_data, self.group, self.options, mvr=self.plot.reg)

    def redraw_plot(self):
        self.plot.redraw_plot()

    def on_resize(self, *evt):
        try:
            self.Refresh()
            self.Layout()
            wx.CallAfter(self.redraw_plot)
        except RuntimeError:
            pass

    def close_ml_frames(self):
        for frame in self.ml_frames:
            if hasattr(frame, 'Close'):
                try:
                    frame.Close()
                except RuntimeError:
                    pass

    def apply_plot_options(self):
        self.plot.apply_options()
        self.plot.redraw_plot()
        for frame in self.ml_frames:
            try:
                frame.plot.apply_options()
                frame.plot.redraw_plot()
            except RuntimeError:
                pass


class LoadMultiVarModelFrame(MultiVarResultsFrame):
    def __init__(self, main_app_frame, model_file_path, group_data, group, options):
        self.loaded_data = load_object_from_file(model_file_path)
        self.stats_data = group_data[group]['stats_data']
        try:
            if self.is_valid:
                y_variable = self.loaded_data['y_variable']
                x_variables = self.loaded_data['x_variables']
                stats_data = group_data[group]['stats_data']

                MultiVarResultsFrame.__init__(self, main_app_frame, y_variable, x_variables, group_data, group, options,
                                              auto_update_plot=False)
                X, y = stats_data.get_X_and_y(y_variable, x_variables)

                reg = MultiVariableRegression(X, y, saved_reg=self.loaded_data['regression'])
                self.plot.update_plot(y_variable, x_variables, self.stats_data, reg=reg)

                self.Show()
            else:
                if self.stats_data is None:
                    msg = 'No data has been queried for Group %s.' % group
                elif not self.is_mvr:
                    msg = 'Selected file is not a valid multi-variable regression save file.'
                elif not self.stats_data_has_y:
                    msg = "The model's dependent variable is not found in your queried data:\n%s" % self.y_variable
                elif self.missing_x_variables:
                    msg = 'Your queried data is missing the following independent variables:\n%s' % \
                          ', '.join(self.missing_x_variables)
                else:
                    msg = 'Unknown error.'

                wx.MessageBox(msg, 'Model Loading Error', wx.OK | wx.OK_DEFAULT | wx.ICON_WARNING)
        except Exception as e:
            msg = str(e)
            wx.MessageBox(msg, 'Model Loading Error', wx.OK | wx.OK_DEFAULT | wx.ICON_WARNING)

    @property
    def is_mvr(self):
        return 'regression_type' in list(self.loaded_data) \
                and self.loaded_data['regression_type'] == 'multi-variable-linear'

    @property
    def is_valid(self):
        return self.stats_data is not None and self.is_mvr and not self.missing_x_variables and self.stats_data_has_y

    @property
    def missing_x_variables(self):
        return [x for x in self.loaded_data['x_variables'] if x not in list(self.stats_data.data)]

    @property
    def stats_data_has_y(self):
        return self.loaded_data['y_variable'] in list(self.stats_data.data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# db.sql_columns.py
"""
These objects are largely used for designing queries in the GUI to easily connect UI friendly variable names with their
associated SQL tables column names
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

# This is a maps categorical data type selections to SQL columns and SQL tables
categorical = {'ROI Institutional Category': {'var_name': 'institutional_roi', 'table': 'DVHs'},
               'ROI Physician Category': {'var_name': 'physician_roi', 'table': 'DVHs'},
               'ROI Name': {'var_name': 'roi_name', 'table': 'DVHs'},
               'ROI Type': {'var_name': 'roi_type', 'table': 'DVHs'},
               'Beam Type': {'var_name': 'beam_type', 'table': 'Beams'},
               'Collimator Rotation Direction': {'var_name': 'collimator_rot_dir', 'table': 'Beams'},
               'Couch Rotation Direction': {'var_name': 'couch_rot_dir', 'table': 'Beams'},
               'Dose Grid Resolution': {'var_name': 'dose_grid_res', 'table': 'Plans'},
               'Gantry Rotation Direction': {'var_name': 'gantry_rot_dir', 'table': 'Beams'},
               'Radiation Type': {'var_name': 'radiation_type', 'table': 'Beams'},
               'Patient Orientation': {'var_name': 'patient_orientation', 'table': 'Plans'},
               'Patient Sex': {'var_name': 'patient_sex', 'table': 'Plans'},
               'Physician': {'var_name': 'physician', 'table': 'Plans'},
               'Tx Modality': {'var_name': 'tx_modality', 'table': 'Plans'},
               'Tx Site': {'var_name': 'tx_site', 'table': 'Plans'},
               'Normalization': {'var_name': 'normalization_method', 'table': 'Rxs'},
               'Treatment Machine': {'var_name': 'treatment_machine', 'table': 'Beams'},
               'Heterogeneity Correction': {'var_name': 'heterogeneity_correction', 'table': 'Plans'},
               'Scan Mode': {'var_name': 'scan_mode', 'table': 'Beams'},
               'MRN': {'var_name': 'mrn', 'table': 'Plans'},
               'UID': {'var_name': 'study_instance_uid', 'table': 'Plans'},
               'Baseline': {'var_name': 'baseline', 'table': 'Plans'},
               'Protocol': {'var_name': 'protocol', 'table': 'Plans'}}

# This is a maps quantitative data type selections to SQL columns and SQL tables, and the bokeh source
numerical = {'Age': {'var_name': 'age', 'table': 'Plans', 'units': ''},
             'Beam Energy Min': {'var_name': 'beam_energy_min', 'table': 'Beams', 'units': ''},
             'Beam Energy Max': {'var_name': 'beam_energy_max', 'table': 'Beams', 'units': ''},
             'Birth Date': {'var_name': 'birth_date', 'table': 'Plans', 'units': ''},
             'Planned Fractions': {'var_name': 'fxs', 'table': 'Plans', 'units': ''},
             'Rx Dose': {'var_name': 'rx_dose', 'table': 'Plans', 'units': 'Gy'},
             'Rx Isodose': {'var_name': 'rx_percent', 'table': 'Rxs', 'units': '%'},
             'Simulation Date': {'var_name': 'sim_study_date', 'table': 'Plans', 'units': ''},
             'Total Plan MU': {'var_name': 'total_mu', 'table': 'Plans', 'units': 'MU'},
             'Fraction Dose': {'var_name': 'fx_dose', 'table': 'Rxs', 'units': 'Gy'},
             'Beam Dose': {'var_name': 'beam_dose', 'table': 'Beams', 'units': 'Gy'},
             'Beam MU': {'var_name': 'beam_mu', 'table': 'Beams', 'units': ''},
             'Control Point Count': {'var_name': 'control_point_count', 'table': 'Beams', 'units': ''},
             'SSD': {'var_name': 'ssd', 'table': 'Beams', 'units': 'cm'},
             'ROI Min Dose': {'var_name': 'min_dose', 'table': 'DVHs', 'units': 'Gy'},
             'ROI Mean Dose': {'var_name': 'mean_dose', 'table': 'DVHs', 'units': 'Gy'},
             'ROI Max Dose': {'var_name': 'max_dose', 'table': 'DVHs', 'units': 'Gy'},
             'ROI Volume': {'var_name': 'volume', 'table': 'DVHs', 'units': 'cm³'},
             'ROI Surface Area': {'var_name': 'surface_area', 'table': 'DVHs', 'units': 'cm²'},
             'ROI Spread X': {'var_name': 'spread_x', 'table': 'DVHs', 'units': 'cm'},
             'ROI Spread Y': {'var_name': 'spread_y', 'table': 'DVHs', 'units': 'cm'},
             'ROI Spread Z': {'var_name': 'spread_z', 'table': 'DVHs', 'units': 'cm'},
             'PTV Distance (Min)': {'var_name': 'dist_to_ptv_min', 'table': 'DVHs', 'units': 'cm'},
             'PTV Distance (Mean)': {'var_name': 'dist_to_ptv_mean', 'table': 'DVHs', 'units': 'cm'},
             'PTV Distance (Median)': {'var_name': 'dist_to_ptv_median', 'table': 'DVHs', 'units': 'cm'},
             'PTV Distance (Max)': {'var_name': 'dist_to_ptv_max', 'table': 'DVHs', 'units': 'cm'},
             'PTV Distance (Centroids)': {'var_name': 'dist_to_ptv_centroids', 'table': 'DVHs', 'units': 'cm'},
             'PTV Overlap': {'var_name': 'ptv_overlap', 'table': 'DVHs', 'units': 'cm³'},
             'Scan Spots': {'var_name': 'scan_spot_count', 'table': 'Beams', 'units': ''},
             'Beam MU per deg': {'var_name': 'beam_mu_per_deg', 'table': 'Beams', 'units': ''},
             'Beam MU per control point': {'var_name': 'beam_mu_per_cp', 'table': 'Beams', 'units': ''},
             'ROI Cross-Section Max': {'var_name': 'cross_section_max', 'table': 'DVHs', 'units': 'cm²'},
             'ROI Cross-Section Median': {'var_name': 'cross_section_median', 'table': 'DVHs', 'units': 'cm²'},
             'Toxicity Grade': {'var_name': 'toxicity_grade', 'table': 'DVHs', 'units': ''},
             'Plan Complexity': {'var_name': 'complexity', 'table': 'Plans', 'units': ''},
             'Beam Complexity (Min)': {'var_name': 'complexity_min', 'table': 'Beams', 'units': ''},
             'Beam Complexity (Mean)': {'var_name': 'complexity_mean', 'table': 'Beams', 'units': ''},
             'Beam Complexity (Median)': {'var_name': 'complexity_median', 'table': 'Beams', 'units': ''},
             'Beam Complexity (Max)': {'var_name': 'complexity_max', 'table': 'Beams', 'units': ''},
             'Beam Area (Min)': {'var_name': 'area_min', 'table': 'Beams', 'units': 'cm²'},
             'Beam Area (Mean)': {'var_name': 'area_mean', 'table': 'Beams', 'units': 'cm²'},
             'Beam Area (Median)': {'var_name': 'area_median', 'table': 'Beams', 'units': 'cm²'},
             'Beam Area (Max)': {'var_name': 'area_max', 'table': 'Beams', 'units': 'cm²'},
             'Beam Perimeter X (Min)': {'var_name': 'x_perim_min', 'table': 'Beams', 'units': 'cm'},
             'Beam Perimeter X (Mean)': {'var_name': 'x_perim_mean', 'table': 'Beams', 'units': 'cm'},
             'Beam Perimeter X (Median)': {'var_name': 'x_perim_median', 'table': 'Beams', 'units': 'cm'},
             'Beam Perimeter X (Max)': {'var_name': 'x_perim_max', 'table': 'Beams', 'units': 'cm'},
             'Beam Perimeter Y (Min)': {'var_name': 'y_perim_min', 'table': 'Beams', 'units': 'cm'},
             'Beam Perimeter Y (Mean)': {'var_name': 'y_perim_mean', 'table': 'Beams', 'units': 'cm'},
             'Beam Perimeter Y (Median)': {'var_name': 'y_perim_median', 'table': 'Beams', 'units': 'cm'},
             'Beam Perimeter Y (Max)': {'var_name': 'y_perim_max', 'table': 'Beams', 'units': 'cm'},
             'Beam Perimeter (Min)': {'var_name': 'perim_min', 'table': 'Beams', 'units': 'cm'},
             'Beam Perimeter (Mean)': {'var_name': 'perim_mean', 'table': 'Beams', 'units': 'cm'},
             'Beam Perimeter (Median)': {'var_name': 'perim_median', 'table': 'Beams', 'units': 'cm'},
             'Beam Perimeter (Max)': {'var_name': 'perim_max', 'table': 'Beams', 'units': 'cm'},
             'Control Point MU (Min)': {'var_name': 'cp_mu_min', 'table': 'Beams', 'units': ''},
             'Control Point MU (Mean)': {'var_name': 'cp_mu_mean', 'table': 'Beams', 'units': ''},
             'Control Point MU (Median)': {'var_name': 'cp_mu_median', 'table': 'Beams', 'units': ''},
             'Control Point MU (Max)': {'var_name': 'cp_mu_max', 'table': 'Beams', 'units': ''},
             'PTV Cross-Section Max': {'var_name': 'ptv_cross_section_max', 'table': 'Plans', 'units': 'cm²'},
             'PTV Cross-Section Median': {'var_name': 'ptv_cross_section_median', 'table': 'Plans', 'units': 'cm²'},
             'PTV Spread X': {'var_name': 'ptv_spread_x', 'table': 'Plans', 'units': 'cm'},
             'PTV Spread Y': {'var_name': 'ptv_spread_y', 'table': 'Plans', 'units': 'cm'},
             'PTV Spread Z': {'var_name': 'ptv_spread_z', 'table': 'Plans', 'units': 'cm'},
             'PTV Surface Area': {'var_name': 'ptv_surface_area', 'table': 'Plans', 'units': 'cm²'},
             'PTV Volume': {'var_name': 'ptv_volume', 'table': 'Plans', 'units': 'cm³'},
             'PTV Max Dose': {'var_name': 'ptv_max_dose', 'table': 'Plans', 'units': 'Gy'},
             'PTV Min Dose': {'var_name': 'ptv_min_dose', 'table': 'Plans', 'units': 'Gy'}}

# Removed these variables from UI as they may be overkill?
numerical_detailed = {'Collimator Start Angle': {'var_name': 'collimator_start', 'table': 'Beams', 'units': 'deg'},
                      'Collimator End Angle': {'var_name': 'collimator_end', 'table': 'Beams', 'units': 'deg'},
                      'Collimator Min Angle': {'var_name': 'collimator_min', 'table': 'Beams', 'units': 'deg'},
                      'Collimator Max Angle': {'var_name': 'collimator_max', 'table': 'Beams', 'units': 'deg'},
                      'Collimator Range': {'var_name': 'collimator_range', 'table': 'Beams', 'units': 'deg'},
                      'Couch Start Angle': {'var_name': 'couch_start', 'table': 'Beams', 'units': 'deg'},
                      'Couch End Angle': {'var_name': 'couch_end', 'table': 'Beams', 'units': 'deg'},
                      'Couch Min Angle': {'var_name': 'couch_min', 'table': 'Beams', 'units': 'deg'},
                      'Couch Max Angle': {'var_name': 'couch_max', 'table': 'Beams', 'units': 'deg'},
                      'Couch Range': {'var_name': 'couch_range', 'table': 'Beams', 'units': 'deg'},
                      'Gantry Start Angle': {'var_name': 'gantry_start', 'table': 'Beams', 'units': 'deg'},
                      'Gantry End Angle': {'var_name': 'gantry_end', 'table': 'Beams', 'units': 'deg'},
                      'Gantry Min Angle': {'var_name': 'gantry_min', 'table': 'Beams', 'units': 'deg'},
                      'Gantry Max Angle': {'var_name': 'gantry_max', 'table': 'Beams', 'units': 'deg'},
                      'Gantry Range': {'var_name': 'gantry_range', 'table': 'Beams', 'units': 'deg'}}
numerical_detailed.update(numerical)

all_columns = {}
all_columns.update(categorical)
all_columns.update(numerical_detailed)
